"""This code defines a dataclass for default input and a dictionary called
ALL_OUT_PARAMS that contains information about various output parameters
related to human body properties, heat exchange, and environmental conditions.

It also includes a function called show_outparam_docs() that generates a
formatted string with the documentation of the output parameters.

The show_outparam_docs() function uses text wrapping to create a
readable documentation string for both regular output parameters and
extra output parameters.

It sorts the parameters alphabetically by key and formats each line with
the parameter's name, meaning, and unit. The resulting documentation
string can be displayed or printed for user reference.
"""

import re
import textwrap
from dataclasses import dataclass
from typing import ClassVar

import numpy as np

# import pandas as pd
from pythermalcomfort.classes_return import JOS3BodyParts
from pythermalcomfort.utilities import BodySurfaceAreaEquations, Postures, Sex


@dataclass(frozen=True)
class Default:
    # Body information
    height: float = 1.72  # [m]
    weight: float = 74.43  # [kg]
    age: int = 20  # [-]
    body_fat: float = 15  # [%]
    cardiac_index: float = 2.59  # [L/min/m2]
    blood_flow_rate: int = 290  # [L/h]
    physical_activity_ratio: float = 1.25  # [-]
    metabolic_rate: float = 1.0  # [met]
    sex: str = Sex.male.value
    posture: str = Postures.standing.value
    bmr_equation: str = "harris-benedict"
    bsa_equation: str = BodySurfaceAreaEquations.dubois.value
    local_bsa: ClassVar[list[float]] = np.array(  # body surface area [m2]
        [
            0.110,
            0.029,
            0.175,
            0.161,
            0.221,
            0.096,
            0.063,
            0.050,
            0.096,
            0.063,
            0.050,
            0.209,
            0.112,
            0.056,
            0.209,
            0.112,
            0.056,
        ],
    )

    # Environment information
    core_temperature: float = 37  # [°C]
    skin_temperature: float = 34  # [°C]
    other_body_temperature: float = 36  # [°C]
    dry_bulb_air_temperature: float = 28.8  # [°C]
    mean_radiant_temperature: float = 28.8  # [°C]
    relative_humidity: float = 50  # [%]
    air_speed: float = 0.1  # [m/s]
    # Clothing information
    clothing_insulation: float = 0  # [clo]
    clothing_vapor_permeation_efficiency: float = 0.45  # [-]
    lewis_rate: float = 16.5  # [K/kPa]
    num_body_parts: int = 17  # [-]


ALL_OUT_PARAMS = {
    "age": {"meaning": "age", "suffix": None, "unit": "years"},
    "bf_ava_foot": {
        "meaning": "AVA blood flow rate of one foot",
        "suffix": None,
        "unit": "L/h",
    },
    "bf_ava_hand": {
        "meaning": "AVA blood flow rate of one hand",
        "suffix": None,
        "unit": "L/h",
    },
    "bf_core": {
        "meaning": "core blood flow rate (each body part)",
        "suffix": "Body name",
        "unit": "L/h",
    },
    "bf_fat": {
        "meaning": "fat blood flow rate (each body part)",
        "suffix": "Body name",
        "unit": "L/h",
    },
    "bf_muscle": {
        "meaning": "muscle blood flow rate (each body part)",
        "suffix": "Body name",
        "unit": "L/h",
    },
    "bf_skin": {
        "meaning": "skin blood flow rate (each body part)",
        "suffix": "Body name",
        "unit": "L/h",
    },
    "bsa": {
        "meaning": "body surface area (each body part)",
        "suffix": "Body name",
        "unit": "m2",
    },
    "cardiac_output": {
        "meaning": "cardiac output (the sum of the whole blood flow)",
        "suffix": None,
        "unit": "L/h",
    },
    "e_max": {
        "meaning": "maximum evaporative heat loss from the skin (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "e_skin": {
        "meaning": "evaporative heat loss from the skin (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "e_sweat": {
        "meaning": (
            "evaporative heat loss from the skin by only sweating (each body part)"
        ),
        "suffix": "Body name",
        "unit": "W",
    },
    "fat": {"meaning": "body fat rate", "suffix": None, "unit": "%"},
    "height": {
        "meaning": "body height",
        "suffix": None,
        "unit": "m",
    },
    "clo": {
        "meaning": "clothing insulation (each body part)",
        "suffix": "Body name",
        "unit": "clo",
    },
    "q_skin2env_latent": {
        "meaning": "latent heat loss from the skin (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "q_bmr_core": {
        "meaning": "core thermogenesis by basal metabolism (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "q_bmr_fat": {
        "meaning": "fat thermogenesis by basal metabolism (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "q_bmr_muscle": {
        "meaning": "muscle thermogenesis by basal metabolism (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "q_bmr_skin": {
        "meaning": "skin thermogenesis by basal metabolism (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "q_thermogenesis_total": {
        "meaning": "total thermogenesis of the whole body",
        "suffix": None,
        "unit": "W",
    },
    "q_nst": {
        "meaning": "core thermogenesis by non-shivering (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "simulation_time": {
        "meaning": "simulation times",
        "suffix": None,
        "unit": "sec",
    },
    "q_shiv": {
        "meaning": "core or muscle thermogenesis by shivering (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "q_work": {
        "meaning": "core or muscle thermogenesis by work (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "name": {
        "meaning": "name of the model",
        "suffix": None,
        "unit": "-",
    },
    "par": {
        "meaning": "physical activity ratio",
        "suffix": None,
        "unit": "-",
    },
    "q_thermogenesis_core": {
        "meaning": "core total thermogenesis (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "q_thermogenesis_fat": {
        "meaning": "fat total thermogenesis (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "q_thermogenesis_muscle": {
        "meaning": "muscle total thermogenesis (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "q_thermogenesis_skin": {
        "meaning": "skin total thermogenesis (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "q_res": {
        "meaning": "heat loss by respiration",
        "suffix": None,
        "unit": "W",
    },
    "q_res_latent": {
        "meaning": "latent heat loss by respiration (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "q_res_sensible": {
        "meaning": "sensible heat loss by respiration (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "rh": {
        "meaning": "relative humidity (each body part)",
        "suffix": "Body name",
        "unit": "%",
    },
    "r_et": {
        "meaning": "total clothing evaporative heat resistance (each body part)",
        "suffix": "Body name",
        "unit": "(m2*kPa)/W",
    },
    "r_t": {
        "meaning": "total clothing heat resistance (each body part)",
        "suffix": "Body name",
        "unit": "(m2*K)/W",
    },
    "q_skin2env_sensible": {
        "meaning": "sensible heat loss from the skin (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "t_skin_set": {
        "meaning": "skin set point temperature (each body part)",
        "suffix": "Body name",
        "unit": "°C",
    },
    "t_core_set": {
        "meaning": "core set point temperature (each body part)",
        "suffix": "Body name",
        "unit": "°C",
    },
    "sex": {
        "meaning": "sex",
        "suffix": None,
        "unit": "-",
    },
    "q_skin2env": {
        "meaning": "total heat loss from the skin (each body part)",
        "suffix": "Body name",
        "unit": "W",
    },
    "tdb": {
        "meaning": "dry bulb air temperature (each body part)",
        "suffix": "Body name",
        "unit": "°C",
    },
    "t_artery": {
        "meaning": "arterial temperature (each body part)",
        "suffix": "Body name",
        "unit": "°C",
    },
    "t_cb": {
        "meaning": "central blood temperature",
        "suffix": None,
        "unit": "°C",
    },
    "t_core": {
        "meaning": "core temperature (each body part)",
        "suffix": "Body name",
        "unit": "°C",
    },
    "t_fat": {
        "meaning": "fat temperature (each body part)",
        "suffix": "Body name",
        "unit": "°C",
    },
    "t_muscle": {
        "meaning": "muscle temperature (each body part)",
        "suffix": "Body name",
        "unit": "°C",
    },
    "to": {
        "meaning": "operative temperature (each body part)",
        "suffix": "Body name",
        "unit": "°C",
    },
    "tr": {
        "meaning": "mean radiant temperature (each body part)",
        "suffix": "Body name",
        "unit": "°C",
    },
    "t_skin": {
        "meaning": "skin temperature (each body part)",
        "suffix": "Body name",
        "unit": "°C",
    },
    "t_skin_mean": {
        "meaning": "mean skin temperature",
        "suffix": None,
        "unit": "°C",
    },
    "t_superficial_vein": {
        "meaning": "superficial vein temperature (each body part)",
        "suffix": "Body name",
        "unit": "°C",
    },
    "t_vein": {
        "meaning": "vein temperature (each body part)",
        "suffix": "Body name",
        "unit": "°C",
    },
    "v": {
        "meaning": "air velocity (each body part)",
        "suffix": "Body name",
        "unit": "m/s",
    },
    "weight": {
        "meaning": "body weight",
        "suffix": None,
        "unit": "kg",
    },
    "w": {
        "meaning": "skin wettedness (each body part)",
        "suffix": "Body name",
        "unit": "-",
    },
    "w_mean": {
        "meaning": "mean skin wettedness",
        "suffix": None,
        "unit": "-",
    },
    "weight_loss_by_evap_and_res": {
        "meaning": "weight loss by the evaporation and respiration of the whole body",
        "suffix": None,
        "unit": "g/sec",
    },
    "dt": {
        "meaning": "time step",
        "suffix": None,
        "unit": "sec",
    },
    "pythermalcomfort_version": {
        "meaning": "version of pythermalcomfort",
        "suffix": None,
        "unit": "-",
    },
}


def show_out_param_docs():
    """Show the documentation of the output parameters.

    Returns
    -------
    docstring : str
        Text of the documentation of the output parameters

    """
    outparams = textwrap.dedent(
        """
        Output parameters
        -------
        """,
    )

    exoutparams = textwrap.dedent(
        """
        Extra output parameters
        -------
        """,
    )

    sortkeys = list(ALL_OUT_PARAMS.keys())
    sortkeys.sort()
    for key in sortkeys:
        value = ALL_OUT_PARAMS[key]

        line = "{}: {} [{}]".format(key.ljust(8), value["meaning"], value["unit"])

        outparams += line + "\n"

    docs = outparams + "\n" + exoutparams
    docs = textwrap.indent(docs.strip(), "    ")

    return docs


# commented out this function since needs pandas
# def convert_and_print_local_clo_values_from_csv_to_dict(csv_name):
#     """Parameters
#     ----------
#     csv_name : file path that you want to convert (it should be in the same folder as this function)
#
#     Returns
#     -------
#     local_clo_dict : a dictionary including local clothing insulation values as well as that for the whole body
#
#     Notes
#     -----
#     The 17 body segments correspond to the JOS-3 model.
#
#     """
#     # Read the Excel file
#     df = pd.read_csv(csv_name)
#
#     # Create an empty dictionary
#     local_clo_dict = {}
#
#     # Add data from each row to the dictionary
#     for _index, row in df.iterrows():
#         # Get the name of the clothing combination as the key
#         key = row["clothing_ensemble"]
#         # Create a dictionary as the value
#         value = {"whole_body": row["whole_body"], "local_body_part": {}}
#         # Add data from each column to the Local body dictionary
#         for col in df.columns[2:]:
#             # Get the name of the body part
#             body_part = col
#             # Get the clo value for the local body part and add it to the Local body dictionary
#             clo = row[col]
#             value["local_body_part"][body_part] = clo
#         # Add the data to the dictionary
#         local_clo_dict[key] = value
#
#     return local_clo_dict


def add_prompt_to_code(code: str, prompt: str = ">>> ") -> str:
    lines = code.strip().split("\n")
    result = []
    for line in lines:
        if re.match(r"^\s*#", line):  # If it's a comment line
            result.append(line)
        else:
            result.append(prompt + line)
    return "\n".join(result)


# This dictionary contains the local and the whole body clothing insulation of typical clothing ensemble.
# It is based on the study by Juyoun et al. (https://escholarship.org/uc/item/18f0r375)
# and by Nomoto et al. (https://doi.org/10.1002/2475-8876.12124)
# Please note that value for the neck is the same as the measured value for the head
# and it does not take into account the insulation effect of the hair.
# Typically, the clothing insulation for the hair are quantified by assuming a head covering of approximately 0.6 to 1.0 clo.

local_clo_typical_ensembles = {
    "nude (mesh chair)": {
        "whole_body": 0.01,
        "local_body_part": JOS3BodyParts(
            head=0.13,
            neck=0.13,
            chest=0.01,
            back=0.01,
            pelvis=0.04,
            left_shoulder=0.02,
            left_arm=0.0,
            left_hand=0.01,
            right_shoulder=0.02,
            right_arm=0.0,
            right_hand=0.01,
            left_thigh=0.01,
            left_leg=0.03,
            left_foot=0.05,
            right_thigh=0.01,
            right_leg=0.03,
            right_foot=0.05,
        ),
    },
    "nude (nude chair)": {
        "whole_body": -0.02,
        "local_body_part": JOS3BodyParts(
            head=0.13,
            neck=0.13,
            chest=0.05,
            back=-0.14,
            pelvis=-0.01,
            left_shoulder=-0.01,
            left_arm=-0.01,
            left_hand=-0.02,
            right_shoulder=-0.01,
            right_arm=-0.01,
            right_hand=-0.02,
            left_thigh=-0.1,
            left_leg=0.0,
            left_foot=0.0,
            right_thigh=-0.1,
            right_leg=0.0,
            right_foot=0.0,
        ),
    },
    "panty": {
        "whole_body": 0.03,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=0.0,
            back=0.0,
            pelvis=0.24,
            left_shoulder=0.0,
            left_arm=0.0,
            left_hand=0.0,
            right_shoulder=0.0,
            right_arm=0.0,
            right_hand=0.0,
            left_thigh=0.05,
            left_leg=0.0,
            left_foot=0.05,
            right_thigh=0.05,
            right_leg=0.0,
            right_foot=0.05,
        ),
    },
    "bra+panty": {
        "whole_body": 0.05,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=0.22,
            back=0.0,
            pelvis=0.18,
            left_shoulder=0.0,
            left_arm=0.0,
            left_hand=0.0,
            right_shoulder=0.0,
            right_arm=0.0,
            right_hand=0.0,
            left_thigh=0.03,
            left_leg=0.03,
            left_foot=0.08,
            right_thigh=0.03,
            right_leg=0.03,
            right_foot=0.08,
        ),
    },
    "bra+panty, tanktop, shorts, sandals": {
        "whole_body": 0.22,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=0.57,
            back=0.27,
            pelvis=0.92,
            left_shoulder=0.04,
            left_arm=0.02,
            left_hand=0.02,
            right_shoulder=0.04,
            right_arm=0.02,
            right_hand=0.02,
            left_thigh=0.51,
            left_leg=0.01,
            left_foot=0.38,
            right_thigh=0.51,
            right_leg=0.01,
            right_foot=0.38,
        ),
    },
    "bra+panty, long-sleeve shirt, shorts, sandals": {
        "whole_body": 0.43,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=1.43,
            back=1.02,
            pelvis=1.45,
            left_shoulder=0.29,
            left_arm=0.22,
            left_hand=0.01,
            right_shoulder=0.29,
            right_arm=0.22,
            right_hand=0.01,
            left_thigh=0.57,
            left_leg=0.01,
            left_foot=0.4,
            right_thigh=0.57,
            right_leg=0.01,
            right_foot=0.4,
        ),
    },
    "bra+panty, sleeveless dress, sandals": {
        "whole_body": 0.29,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=0.85,
            back=0.48,
            pelvis=0.94,
            left_shoulder=0.0,
            left_arm=0.0,
            left_hand=0.0,
            right_shoulder=0.0,
            right_arm=0.0,
            right_hand=0.0,
            left_thigh=0.72,
            left_leg=0.0,
            left_foot=0.41,
            right_thigh=0.72,
            right_leg=0.0,
            right_foot=0.41,
        ),
    },
    "bra+panty, T-shirt, long pants, socks, sneakers": {
        "whole_body": 0.52,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=1.14,
            back=0.84,
            pelvis=1.04,
            left_shoulder=0.42,
            left_arm=0.0,
            left_hand=0.0,
            right_shoulder=0.42,
            right_arm=0.0,
            right_hand=0.0,
            left_thigh=0.58,
            left_leg=0.62,
            left_foot=0.82,
            right_thigh=0.58,
            right_leg=0.62,
            right_foot=0.82,
        ),
    },
    "bra+panty, sleeveless dress, cardigan, sandals": {
        "whole_body": 0.53,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=1.78,
            back=1.42,
            pelvis=1.19,
            left_shoulder=0.65,
            left_arm=0.41,
            left_hand=0.05,
            right_shoulder=0.65,
            right_arm=0.41,
            right_hand=0.05,
            left_thigh=0.77,
            left_leg=0.0,
            left_foot=0.39,
            right_thigh=0.77,
            right_leg=0.0,
            right_foot=0.39,
        ),
    },
    "bra+panty, song-sleeve dress, socks, sneakers": {
        "whole_body": 0.54,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=1.49,
            back=1.1,
            pelvis=0.91,
            left_shoulder=0.72,
            left_arm=0.58,
            left_hand=0.03,
            right_shoulder=0.72,
            right_arm=0.58,
            right_hand=0.03,
            left_thigh=0.73,
            left_leg=0.07,
            left_foot=0.77,
            right_thigh=0.73,
            right_leg=0.07,
            right_foot=0.77,
        ),
    },
    "bra+panty, long-sleeve dress, cardigan, socks, sneakers": {
        "whole_body": 0.67,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=2.05,
            back=1.32,
            pelvis=1.39,
            left_shoulder=1.14,
            left_arm=0.63,
            left_hand=0.04,
            right_shoulder=1.14,
            right_arm=0.63,
            right_hand=0.04,
            left_thigh=0.84,
            left_leg=0.05,
            left_foot=0.78,
            right_thigh=0.84,
            right_leg=0.05,
            right_foot=0.78,
        ),
    },
    "bra+panty, tank top, skirt, sandals": {
        "whole_body": 0.31,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=0.83,
            back=0.22,
            pelvis=0.99,
            left_shoulder=0.0,
            left_arm=0.0,
            left_hand=0.03,
            right_shoulder=0.0,
            right_arm=0.0,
            right_hand=0.03,
            left_thigh=0.88,
            left_leg=0.05,
            left_foot=0.44,
            right_thigh=0.88,
            right_leg=0.05,
            right_foot=0.44,
        ),
    },
    "bra+panty, long sleeve shirts, skirt, sandals": {
        "whole_body": 0.52,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=1.62,
            back=0.99,
            pelvis=1.41,
            left_shoulder=0.31,
            left_arm=0.28,
            left_hand=0.03,
            right_shoulder=0.31,
            right_arm=0.28,
            right_hand=0.03,
            left_thigh=0.82,
            left_leg=0.04,
            left_foot=0.41,
            right_thigh=0.82,
            right_leg=0.04,
            right_foot=0.41,
        ),
    },
    "bra+panty, dress shirts, skirt, stocking, formal shoes": {
        "whole_body": 0.62,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=1.58,
            back=0.99,
            pelvis=1.31,
            left_shoulder=0.91,
            left_arm=0.64,
            left_hand=0.04,
            right_shoulder=0.91,
            right_arm=0.64,
            right_hand=0.04,
            left_thigh=0.87,
            left_leg=0.05,
            left_foot=0.81,
            right_thigh=0.87,
            right_leg=0.05,
            right_foot=0.81,
        ),
    },
    "bra+panty, dress shirts, skirt, leggings, sandals": {
        "whole_body": 0.65,
        "local_body_part": JOS3BodyParts(
            head=0.13,
            neck=0.13,
            chest=1.59,
            back=1.04,
            pelvis=1.36,
            left_shoulder=0.91,
            left_arm=0.67,
            left_hand=0.07,
            right_shoulder=0.91,
            right_arm=0.67,
            right_hand=0.07,
            left_thigh=1.26,
            left_leg=0.12,
            left_foot=0.43,
            right_thigh=1.26,
            right_leg=0.12,
            right_foot=0.43,
        ),
    },
    "bra+panty, thin dress shirts, long pants, socks, sneakers": {
        "whole_body": 0.82,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=3.35,
            back=1.73,
            pelvis=1.63,
            left_shoulder=1.99,
            left_arm=1.49,
            left_hand=0.11,
            right_shoulder=1.99,
            right_arm=1.49,
            right_hand=0.11,
            left_thigh=0.6,
            left_leg=0.43,
            left_foot=0.68,
            right_thigh=0.6,
            right_leg=0.43,
            right_foot=0.68,
        ),
    },
    "bra+panty, long sleeve shirts, long pants, socks, sneakers": {
        "whole_body": 0.8,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=2.47,
            back=1.48,
            pelvis=1.58,
            left_shoulder=0.98,
            left_arm=0.58,
            left_hand=0.04,
            right_shoulder=0.98,
            right_arm=0.58,
            right_hand=0.04,
            left_thigh=0.69,
            left_leg=0.65,
            left_foot=0.89,
            right_thigh=0.69,
            right_leg=0.65,
            right_foot=0.89,
        ),
    },
    "bra+panty, T-shirt, long sleeve shirts, long pants, socks, sneakers": {
        "whole_body": 0.83,
        "local_body_part": JOS3BodyParts(
            head=0.25,
            neck=0.25,
            chest=3.88,
            back=2.28,
            pelvis=2.07,
            left_shoulder=1.89,
            left_arm=1.41,
            left_hand=0.16,
            right_shoulder=1.89,
            right_arm=1.41,
            right_hand=0.16,
            left_thigh=0.83,
            left_leg=0.66,
            left_foot=0.86,
            right_thigh=0.83,
            right_leg=0.66,
            right_foot=0.86,
        ),
    },
    "bra+panty, T-shirt, jeans, socks, sneakers": {
        "whole_body": 0.57,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=1.29,
            back=0.93,
            pelvis=1.3,
            left_shoulder=0.68,
            left_arm=0.0,
            left_hand=0.0,
            right_shoulder=0.68,
            right_arm=0.0,
            right_hand=0.0,
            left_thigh=0.65,
            left_leg=0.47,
            left_foot=0.73,
            right_thigh=0.65,
            right_leg=0.47,
            right_foot=0.73,
        ),
    },
    "bra+panty, long sleeve shirts, jeans, socks, sneakers": {
        "whole_body": 0.74,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=1.58,
            back=0.98,
            pelvis=1.35,
            left_shoulder=0.86,
            left_arm=0.71,
            left_hand=0.07,
            right_shoulder=0.86,
            right_arm=0.71,
            right_hand=0.07,
            left_thigh=0.74,
            left_leg=0.48,
            left_foot=0.74,
            right_thigh=0.74,
            right_leg=0.48,
            right_foot=0.74,
        ),
    },
    "bra+panty, oxford shirts, long thin pants, socks, sneakers": {
        "whole_body": 0.83,
        "local_body_part": JOS3BodyParts(
            head=0.16,
            neck=0.16,
            chest=1.39,
            back=1.02,
            pelvis=1.34,
            left_shoulder=0.83,
            left_arm=0.69,
            left_hand=0.22,
            right_shoulder=0.83,
            right_arm=0.69,
            right_hand=0.22,
            left_thigh=1.02,
            left_leg=0.68,
            left_foot=0.8,
            right_thigh=1.02,
            right_leg=0.68,
            right_foot=0.8,
        ),
    },
    "bra+panty, thin dress shirts (roll-up), long pants, socks, sneakers": {
        "whole_body": 0.81,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=3.6,
            back=1.83,
            pelvis=1.71,
            left_shoulder=2.16,
            left_arm=1.49,
            left_hand=0.13,
            right_shoulder=2.16,
            right_arm=1.49,
            right_hand=0.13,
            left_thigh=0.64,
            left_leg=0.43,
            left_foot=0.69,
            right_thigh=0.64,
            right_leg=0.43,
            right_foot=0.69,
        ),
    },
    "bra+panty, T-shirt, short sleeve shirt, long pants, socks, sneakers": {
        "whole_body": 0.71,
        "local_body_part": JOS3BodyParts(
            head=0.12,
            neck=0.12,
            chest=2.15,
            back=1.4,
            pelvis=1.71,
            left_shoulder=1.22,
            left_arm=0.02,
            left_hand=0.05,
            right_shoulder=1.22,
            right_arm=0.02,
            right_hand=0.05,
            left_thigh=0.79,
            left_leg=0.48,
            left_foot=0.67,
            right_thigh=0.79,
            right_leg=0.48,
            right_foot=0.67,
        ),
    },
    "bra+panty, sports shirts, long pants, socks, sneakers": {
        "whole_body": 0.8,
        "local_body_part": JOS3BodyParts(
            head=0.05,
            neck=0.05,
            chest=1.92,
            back=1.31,
            pelvis=1.41,
            left_shoulder=1.14,
            left_arm=0.86,
            left_hand=0.18,
            right_shoulder=1.14,
            right_arm=0.86,
            right_hand=0.18,
            left_thigh=0.59,
            left_leg=0.49,
            left_foot=0.75,
            right_thigh=0.59,
            right_leg=0.49,
            right_foot=0.75,
        ),
    },
    "bra+panty, sports shirts, sports pants, sports socks, sports shoes": {
        "whole_body": 0.87,
        "local_body_part": JOS3BodyParts(
            head=0.07,
            neck=0.07,
            chest=1.87,
            back=1.17,
            pelvis=1.26,
            left_shoulder=1.2,
            left_arm=1.07,
            left_hand=0.09,
            right_shoulder=1.2,
            right_arm=1.07,
            right_hand=0.09,
            left_thigh=0.62,
            left_leg=0.77,
            left_foot=1.58,
            right_thigh=0.62,
            right_leg=0.77,
            right_foot=1.58,
        ),
    },
    "bra+panty, thin dress shirts, long pants, wool sweater, socks, sneakers": {
        "whole_body": 0.92,
        "local_body_part": JOS3BodyParts(
            head=0.09,
            neck=0.09,
            chest=2.39,
            back=1.64,
            pelvis=1.71,
            left_shoulder=1.36,
            left_arm=1.29,
            left_hand=0.21,
            right_shoulder=1.36,
            right_arm=1.29,
            right_hand=0.21,
            left_thigh=0.7,
            left_leg=0.52,
            left_foot=0.77,
            right_thigh=0.7,
            right_leg=0.52,
            right_foot=0.77,
        ),
    },
    "bra+panty, thin dress shirts, long pants, cashmere sweater, socks, sneakers": {
        "whole_body": 0.87,
        "local_body_part": JOS3BodyParts(
            head=0.1,
            neck=0.1,
            chest=2.4,
            back=1.72,
            pelvis=1.67,
            left_shoulder=1.33,
            left_arm=1.23,
            left_hand=0.08,
            right_shoulder=1.33,
            right_arm=1.23,
            right_hand=0.08,
            left_thigh=0.61,
            left_leg=0.47,
            left_foot=0.77,
            right_thigh=0.61,
            right_leg=0.47,
            right_foot=0.77,
        ),
    },
    "bra+panty, T-shirt, long sleeve shirts, long pants, winter jacket, socks, sneakers": {
        "whole_body": 1.18,
        "local_body_part": JOS3BodyParts(
            head=0.65,
            neck=0.65,
            chest=5.26,
            back=3.07,
            pelvis=2.2,
            left_shoulder=3.14,
            left_arm=2.07,
            left_hand=0.08,
            right_shoulder=3.14,
            right_arm=2.07,
            right_hand=0.08,
            left_thigh=0.67,
            left_leg=0.54,
            left_foot=0.77,
            right_thigh=0.67,
            right_leg=0.54,
            right_foot=0.77,
        ),
    },
    "bra+panty, T-shirt, long sleeve shirts, jeans, sports jumper, socks, sneakers": {
        "whole_body": 1.07,
        "local_body_part": JOS3BodyParts(
            head=0.28,
            neck=0.28,
            chest=3.99,
            back=2.12,
            pelvis=2.0,
            left_shoulder=1.7,
            left_arm=1.36,
            left_hand=0.1,
            right_shoulder=1.7,
            right_arm=1.36,
            right_hand=0.1,
            left_thigh=0.92,
            left_leg=0.48,
            left_foot=1.07,
            right_thigh=0.92,
            right_leg=0.48,
            right_foot=1.07,
        ),
    },
    "bra+panty, T-shirt, long sleeve shirts, long pants, ventura jacket, socks, sneakers": {
        "whole_body": 0.9,
        "local_body_part": JOS3BodyParts(
            head=0.09,
            neck=0.09,
            chest=2.66,
            back=1.42,
            pelvis=1.57,
            left_shoulder=1.32,
            left_arm=0.99,
            left_hand=0.14,
            right_shoulder=1.32,
            right_arm=0.99,
            right_hand=0.14,
            left_thigh=0.73,
            left_leg=0.66,
            left_foot=0.85,
            right_thigh=0.73,
            right_leg=0.66,
            right_foot=0.85,
        ),
    },
    "bra+panty, turtle neck, long pants, short trench coat, socks, sneakers": {
        "whole_body": 1.24,
        "local_body_part": JOS3BodyParts(
            head=0.06,
            neck=0.06,
            chest=3.22,
            back=1.99,
            pelvis=2.03,
            left_shoulder=1.62,
            left_arm=1.5,
            left_hand=0.37,
            right_shoulder=1.62,
            right_arm=1.5,
            right_hand=0.37,
            left_thigh=1.51,
            left_leg=0.65,
            left_foot=0.8,
            right_thigh=1.51,
            right_leg=0.65,
            right_foot=0.8,
        ),
    },
    "bra+panty, tank top, long sleeve shirts, blazer, skirt, sandals": {
        "whole_body": 0.86,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=3.24,
            back=1.81,
            pelvis=2.06,
            left_shoulder=1.98,
            left_arm=1.13,
            left_hand=0.07,
            right_shoulder=1.98,
            right_arm=1.13,
            right_hand=0.07,
            left_thigh=1.19,
            left_leg=0.04,
            left_foot=0.44,
            right_thigh=1.19,
            right_leg=0.04,
            right_foot=0.44,
        ),
    },
    "bra+panty, long sleeve shirts, wool skirt, socks, formal shoes": {
        "whole_body": 0.59,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=1.21,
            back=0.74,
            pelvis=1.56,
            left_shoulder=0.44,
            left_arm=0.24,
            left_hand=0.17,
            right_shoulder=0.44,
            right_arm=0.24,
            right_hand=0.17,
            left_thigh=1.52,
            left_leg=0.09,
            left_foot=0.74,
            right_thigh=1.52,
            right_leg=0.09,
            right_foot=0.74,
        ),
    },
    "bra+panty, turtleneck, wool skirt, socks, formal shoes": {
        "whole_body": 0.7,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=1.11,
            back=0.94,
            pelvis=1.52,
            left_shoulder=0.73,
            left_arm=0.62,
            left_hand=0.14,
            right_shoulder=0.73,
            right_arm=0.62,
            right_hand=0.14,
            left_thigh=1.53,
            left_leg=0.09,
            left_foot=0.85,
            right_thigh=1.53,
            right_leg=0.09,
            right_foot=0.85,
        ),
    },
    "bra+panty, long sleeve shirt, wool skirt, sweater, socks, formal shoes": {
        "whole_body": 0.91,
        "local_body_part": JOS3BodyParts(
            head=0.14,
            neck=0.14,
            chest=2.82,
            back=1.53,
            pelvis=1.79,
            left_shoulder=1.22,
            left_arm=0.97,
            left_hand=0.08,
            right_shoulder=1.22,
            right_arm=0.97,
            right_hand=0.08,
            left_thigh=1.53,
            left_leg=0.11,
            left_foot=0.83,
            right_thigh=1.53,
            right_leg=0.11,
            right_foot=0.83,
        ),
    },
    "bra+panty, thin dress shirts, slacks, tie, socks, sneakers": {
        "whole_body": 0.57,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=1.69,
            back=0.8,
            pelvis=1.08,
            left_shoulder=0.67,
            left_arm=0.58,
            left_hand=0.07,
            right_shoulder=0.67,
            right_arm=0.58,
            right_hand=0.07,
            left_thigh=0.36,
            left_leg=0.39,
            left_foot=0.74,
            right_thigh=0.36,
            right_leg=0.39,
            right_foot=0.74,
        ),
    },
    "bra+panty, thin dress shirts, slacks, blazer, tie, belt, socks, formal shoes": {
        "whole_body": 0.93,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=3.6,
            back=1.83,
            pelvis=1.71,
            left_shoulder=2.16,
            left_arm=1.49,
            left_hand=0.13,
            right_shoulder=2.16,
            right_arm=1.49,
            right_hand=0.13,
            left_thigh=0.64,
            left_leg=0.43,
            left_foot=0.69,
            right_thigh=0.64,
            right_leg=0.43,
            right_foot=0.69,
        ),
    },
    "bra+panty, long sleeve shirts, long pants, blazer, socks, sneakers": {
        "whole_body": 0.96,
        "local_body_part": JOS3BodyParts(
            head=0.04,
            neck=0.04,
            chest=3.3,
            back=1.67,
            pelvis=2.2,
            left_shoulder=2.1,
            left_arm=1.43,
            left_hand=0.09,
            right_shoulder=2.1,
            right_arm=1.43,
            right_hand=0.09,
            left_thigh=0.72,
            left_leg=0.42,
            left_foot=0.67,
            right_thigh=0.72,
            right_leg=0.42,
            right_foot=0.67,
        ),
    },
    "bra+panty, T-shirt, long sleeve shirts, long pants, winter jacket (Notica)": {
        "whole_body": 1.05,
        "local_body_part": JOS3BodyParts(
            head=0.04,
            neck=0.04,
            chest=3.88,
            back=2.26,
            pelvis=1.97,
            left_shoulder=1.82,
            left_arm=1.46,
            left_hand=0.17,
            right_shoulder=1.82,
            right_arm=1.46,
            right_hand=0.17,
            left_thigh=0.81,
            left_leg=0.57,
            left_foot=0.78,
            right_thigh=0.81,
            right_leg=0.57,
            right_foot=0.78,
        ),
    },
    "bra+panty, turtle neck, ski-jumper, skin pants, sports socks, sports shoes": {
        "whole_body": 1.84,
        "local_body_part": JOS3BodyParts(
            head=0.89,
            neck=0.89,
            chest=5.24,
            back=2.87,
            pelvis=2.64,
            left_shoulder=2.55,
            left_arm=2.16,
            left_hand=0.46,
            right_shoulder=2.55,
            right_arm=2.16,
            right_hand=0.46,
            left_thigh=1.49,
            left_leg=1.82,
            left_foot=1.56,
            right_thigh=1.49,
            right_leg=1.82,
            right_foot=1.56,
        ),
    },
    "bra+panty, turtle neck, ski-jumper and hood, skin pants, sports socks, sports shoes": {
        "whole_body": 1.87,
        "local_body_part": JOS3BodyParts(
            head=1.63,
            neck=1.63,
            chest=5.12,
            back=2.7,
            pelvis=2.57,
            left_shoulder=2.58,
            left_arm=2.16,
            left_hand=0.49,
            right_shoulder=2.58,
            right_arm=2.16,
            right_hand=0.49,
            left_thigh=1.44,
            left_leg=1.76,
            left_foot=1.54,
            right_thigh=1.44,
            right_leg=1.76,
            right_foot=1.54,
        ),
    },
    "bra+panty, turtle neck, goose down, ski pants, sports socks, sports shoes": {
        "whole_body": 2.53,
        "local_body_part": JOS3BodyParts(
            head=1.17,
            neck=1.17,
            chest=15.44,
            back=5.5,
            pelvis=5.2,
            left_shoulder=6.55,
            left_arm=5.58,
            left_hand=0.35,
            right_shoulder=6.55,
            right_arm=5.58,
            right_hand=0.35,
            left_thigh=2.12,
            left_leg=1.7,
            left_foot=1.54,
            right_thigh=2.12,
            right_leg=1.7,
            right_foot=1.54,
        ),
    },
    "bra+panty, turtle neck, goose down-with hood, ski pants, sports socks, sports shoes": {
        "whole_body": 2.75,
        "local_body_part": JOS3BodyParts(
            head=3.52,
            neck=3.52,
            chest=12.62,
            back=3.99,
            pelvis=5.05,
            left_shoulder=6.2,
            left_arm=5.73,
            left_hand=0.53,
            right_shoulder=6.2,
            right_arm=5.73,
            right_hand=0.53,
            left_thigh=2.11,
            left_leg=1.81,
            left_foot=1.58,
            right_thigh=2.11,
            right_leg=1.81,
            right_foot=1.58,
        ),
    },
    "bra+panty, turtle neck, goose down-with hood and gloves, ski pants, sports socks, sports shoes": {
        "whole_body": 3.27,
        "local_body_part": JOS3BodyParts(
            head=3.92,
            neck=3.92,
            chest=16.13,
            back=4.47,
            pelvis=5.71,
            left_shoulder=7.12,
            left_arm=5.37,
            left_hand=2.54,
            right_shoulder=7.12,
            right_arm=5.37,
            right_hand=2.54,
            left_thigh=2.14,
            left_leg=1.82,
            left_foot=1.61,
            right_thigh=2.14,
            right_leg=1.82,
            right_foot=1.61,
        ),
    },
    "briefs, socks, T-shirt, half pants, sneakers": {
        "whole_body": 0.53,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=0.5,
            back=1.13,
            pelvis=1.21,
            left_shoulder=0.39,
            left_arm=0.0,
            left_hand=0.0,
            right_shoulder=0.39,
            right_arm=0.0,
            right_hand=0.0,
            left_thigh=0.94,
            left_leg=0.07,
            left_foot=0.62,
            right_thigh=0.94,
            right_leg=0.07,
            right_foot=0.62,
        ),
    },
    "briefs, undershirt, sports t-shirts, sports shorts": {
        "whole_body": 0.7,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=0.9,
            back=1.76,
            pelvis=2.14,
            left_shoulder=0.48,
            left_arm=0.0,
            left_hand=0.0,
            right_shoulder=0.48,
            right_arm=0.0,
            right_hand=0.0,
            left_thigh=1.18,
            left_leg=0.0,
            left_foot=0.01,
            right_thigh=1.18,
            right_leg=0.0,
            right_foot=0.01,
        ),
    },
    "briefs, socks, polo shirt, long pants, sneakers": {
        "whole_body": 0.54,
        "local_body_part": JOS3BodyParts(
            head=0.02,
            neck=0.02,
            chest=0.52,
            back=0.97,
            pelvis=1.12,
            left_shoulder=0.34,
            left_arm=0.0,
            left_hand=0.0,
            right_shoulder=0.34,
            right_arm=0.0,
            right_hand=0.0,
            left_thigh=0.79,
            left_leg=0.56,
            left_foot=0.65,
            right_thigh=0.79,
            right_leg=0.56,
            right_foot=0.65,
        ),
    },
    "briefs, under shirt, long-sleeved shirt, long pants": {
        "whole_body": 1.01,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=1.2,
            back=1.65,
            pelvis=2.29,
            left_shoulder=0.98,
            left_arm=0.78,
            left_hand=0.03,
            right_shoulder=0.98,
            right_arm=0.78,
            right_hand=0.03,
            left_thigh=1.46,
            left_leg=0.62,
            left_foot=0.02,
            right_thigh=1.46,
            right_leg=0.62,
            right_foot=0.02,
        ),
    },
    "briefs, socks, undershirt, short-sleeved shirt, long pants, belt, shoes": {
        "whole_body": 0.72,
        "local_body_part": JOS3BodyParts(
            head=0.01,
            neck=0.01,
            chest=0.8,
            back=1.4,
            pelvis=1.55,
            left_shoulder=0.54,
            left_arm=0.0,
            left_hand=0.0,
            right_shoulder=0.54,
            right_arm=0.0,
            right_hand=0.0,
            left_thigh=0.89,
            left_leg=0.64,
            left_foot=0.99,
            right_thigh=0.89,
            right_leg=0.64,
            right_foot=0.99,
        ),
    },
    "briefs, socks, undershirt, long-sleeved shirt, long pants, belt": {
        "whole_body": 0.77,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=0.86,
            back=1.45,
            pelvis=1.54,
            left_shoulder=0.82,
            left_arm=0.6,
            left_hand=0.01,
            right_shoulder=0.82,
            right_arm=0.6,
            right_hand=0.01,
            left_thigh=0.9,
            left_leg=0.66,
            left_foot=0.64,
            right_thigh=0.9,
            right_leg=0.66,
            right_foot=0.64,
        ),
    },
    "briefs, socks, undershirt, long-sleeved shirt, jacket, long pants, belt, shoes": {
        "whole_body": 1.39,
        "local_body_part": JOS3BodyParts(
            head=0.02,
            neck=0.02,
            chest=2.13,
            back=2.28,
            pelvis=3.04,
            left_shoulder=1.8,
            left_arm=1.54,
            left_hand=0.15,
            right_shoulder=1.8,
            right_arm=1.54,
            right_hand=0.15,
            left_thigh=1.33,
            left_leg=0.69,
            left_foot=0.97,
            right_thigh=1.33,
            right_leg=0.69,
            right_foot=0.97,
        ),
    },
    "briefs, socks, undershirt, work jacket, work pants, safety shoes": {
        "whole_body": 0.8,
        "local_body_part": JOS3BodyParts(
            head=0.0,
            neck=0.0,
            chest=1.25,
            back=1.39,
            pelvis=1.78,
            left_shoulder=0.84,
            left_arm=0.71,
            left_hand=0.08,
            right_shoulder=0.84,
            right_arm=0.71,
            right_hand=0.08,
            left_thigh=0.65,
            left_leg=0.59,
            left_foot=1.12,
            right_thigh=0.65,
            right_leg=0.59,
            right_foot=1.12,
        ),
    },
}


if __name__ == "__main__":
    print(show_out_param_docs())



from __future__ import annotations

import math

import numpy as np
from numba import float64, jit, vectorize

from pythermalcomfort.classes_input import GaggeTwoNodesInputs
from pythermalcomfort.classes_return import SET, GaggeTwoNodes
from pythermalcomfort.utilities import Postures, met_to_w_m2, p_sat_torr


def two_nodes_gagge(
    tdb: float | list[float],
    tr: float | list[float],
    v: float | list[float],
    rh: float | list[float],
    met: float | list[float],
    clo: float | list[float],
    wme: float | list[float] = 0,
    body_surface_area: float | list[float] = 1.8258,
    p_atm: float | list[float] = 101325,
    position: str = Postures.standing.value,
    max_skin_blood_flow: float | list[float] = 90,
    round_output: bool = True,
    max_sweating: float | list[float] = 500,
    w_max: float | list[float] = False,
    calculate_ce: bool = False,
) -> SET | GaggeTwoNodes:
    """Gagge Two-node model of human temperature regulation Gagge et al (1986) [Gagge1986]_.

    [Gagge1986]_ This model can be used to calculate a variety of indices, including:

    * Gagge's version of Fanger's Predicted Mean Vote (PMV). This function uses the Fanger's PMV equations but it replaces the heat loss and gain terms with those calculated by the two-node model developed by Gagge et al. (1986) [Gagge1986]_.
    * PMV SET and the predicted thermal sensation based on SET [Gagge1986]_. This function is similar in all aspects to the :py:meth:`pythermalcomfort.models.pmv_gagge` however, it uses the :py:meth:`pythermalcomfort.models.set` equation to calculate the dry heat loss by convection.
    * Thermal discomfort (DISC) as the relative thermoregulatory strain necessary to restore a state of comfort and thermal equilibrium by sweating [Gagge1986]_. DISC is described numerically as: comfortable and pleasant (0), slightly uncomfortable but acceptable (1), uncomfortable and unpleasant (2), very uncomfortable (3), limited tolerance (4), and intolerable (5). The range of each category is ± 0.5 numerically. In the cold, the classical negative category descriptions used for Fanger's PMV apply [Gagge1986]_.
    * Heat gains and losses via convection, radiation, and conduction.
    * The Standard Effective Temperature (SET).
    * The New Effective Temperature (ET).
    * The Predicted Thermal Sensation (TSENS).
    * The Predicted Percent Dissatisfied Due to Draft (PD).
    * Predicted Percent Satisfied With the Level of Air Movement (PS).

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    tr : float or list of floats
        Mean radiant temperature, [°C].
    v : float or list of floats
        Air speed, [m/s].
    rh : float or list of floats
        Relative humidity, [%].
    met : float or list of floats
        Metabolic rate, [met].
    clo : float or list of floats
        Clothing insulation, [clo].
    wme : float or list of floats, optional
        External work, [met]. Defaults to 0.
    body_surface_area : float or list of floats, optional
        Body surface area, default value 1.8258 [m2]. Defaults to 1.8258.

        .. note::
            The body surface area can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.body_surface_area`.

    p_atm : float or list of floats, optional
        Atmospheric pressure, default value 101325 [Pa]. Defaults to 101325.
    position : str, optional
        Select either "sitting" or "standing". Defaults to "standing".
    max_skin_blood_flow : float or list of floats, optional
        Maximum blood flow from the core to the skin, [kg/h/m2]. Defaults to 90.
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.
    max_sweating : float or list of floats, optional
        Maximum rate at which regulatory sweat is generated, [kg/h/m2]. Defaults to 500.
    w_max : float or list of floats, optional
        Maximum skin wettedness (w) adimensional. Ranges from 0 and 1. Defaults to False.

    Returns
    -------
    GaggeTwoNodes
        A dataclass containing the results of the two-node model of human temperature regulation.
        See :py:class:`~pythermalcomfort.classes_return.GaggeTwoNodes` for more details.
        To access the results, use the corresponding attributes of the returned `GaggeTwoNodes` instance, e.g., `result.e_skin`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import two_nodes_gagge

        result = two_nodes_gagge(tdb=25, tr=25, v=0.1, rh=50, clo=0.5, met=1.2)
        print(result.w)  # 100.0

        result = two_nodes_gagge(tdb=[25, 25], tr=25, v=0.3, rh=50, met=1.2, clo=0.5)
        print(result.e_skin)  # [100.0, 100.0]

    """
    # Validate inputs using the TwoNodesInputs class
    GaggeTwoNodesInputs(
        tdb=tdb,
        tr=tr,
        v=v,
        rh=rh,
        met=met,
        clo=clo,
        wme=wme,
        body_surface_area=body_surface_area,
        p_atm=p_atm,
        position=position,
        max_skin_blood_flow=max_skin_blood_flow,
        round_output=round_output,
        max_sweating=max_sweating,
        w_max=w_max,
    )

    tdb = np.array(tdb)
    tr = np.array(tr)
    v = np.array(v)
    rh = np.array(rh)
    met = np.array(met)
    clo = np.array(clo)
    wme = np.array(wme)
    position = np.array(position)

    vapor_pressure = rh * p_sat_torr(tdb) / 100

    if calculate_ce:
        result = _gagge_two_nodes_optimized_return_set(
            tdb,
            tr,
            v,
            met,
            clo,
            vapor_pressure,
            wme,
            body_surface_area,
            p_atm,
            1,
        )
        return SET(set=result)

    (
        _set,
        e_skin,
        e_rsw,
        e_max,
        q_sensible,
        q_skin,
        q_res,
        t_core,
        t_skin,
        m_bl,
        m_rsw,
        w,
        w_max,
        et,
        pmv_gagge,
        pmv_set,
        disc,
        t_sens,
    ) = np.vectorize(_gagge_two_nodes_optimized, cache=True)(
        tdb=tdb,
        tr=tr,
        v=v,
        met=met,
        clo=clo,
        vapor_pressure=vapor_pressure,
        wme=wme,
        body_surface_area=body_surface_area,
        p_atm=p_atm,
        position=position,
        calculate_ce=calculate_ce,
        max_skin_blood_flow=max_skin_blood_flow,
        max_sweating=max_sweating,
        w_max=w_max,
    )

    output = {
        "e_skin": e_skin,
        "e_rsw": e_rsw,
        "e_max": e_max,
        "q_sensible": q_sensible,
        "q_skin": q_skin,
        "q_res": q_res,
        "t_core": t_core,
        "t_skin": t_skin,
        "m_bl": m_bl,
        "m_rsw": m_rsw,
        "w": w,
        "w_max": w_max,
        "set": _set,
        "et": et,
        "pmv_gagge": pmv_gagge,
        "pmv_set": pmv_set,
        "disc": disc,
        "t_sens": t_sens,
    }

    if round_output:
        for key in output:
            output[key] = np.around(output[key], 2)

    return GaggeTwoNodes(**output)


@jit(nopython=True, cache=True)
def _gagge_two_nodes_optimized(
    tdb,
    tr,
    v,
    met,
    clo,
    vapor_pressure,
    wme,
    body_surface_area,
    p_atm,
    position,
    calculate_ce=False,
    max_skin_blood_flow=90,
    max_sweating=500,
    w_max=None,
) -> tuple[
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
]:
    # Initial variables as defined in the ASHRAE 55-2020
    air_speed = max(v, 0.1)
    k_clo = 0.25
    body_weight = 70  # body weight in kg
    met_factor = 58.2  # met conversion factor
    sbc = 0.000000056697  # Stefan-Boltzmann constant (W/m2K4)
    c_sw = 170  # driving coefficient for regulatory sweating
    c_dil = 120  # driving coefficient for vasodilation ashrae says 50 see page 9.19
    c_str = 0.5  # driving coefficient for vasoconstriction

    temp_skin_neutral = 33.7
    temp_core_neutral = 36.8
    alfa = 0.1
    temp_body_neutral = alfa * temp_skin_neutral + (1 - alfa) * temp_core_neutral
    skin_blood_flow_neutral = 6.3

    t_skin = temp_skin_neutral
    t_core = temp_core_neutral
    m_bl = skin_blood_flow_neutral

    # initialize some variables
    e_skin = 0.1 * met  # total evaporative heat loss, W
    q_sensible = 0  # total sensible heat loss, W
    w = 0  # skin wettedness
    _set = 0  # standard effective temperature
    e_rsw = 0  # heat lost by vaporization sweat
    e_diff = 0  # vapor diffusion through skin
    e_max = 0  # maximum evaporative capacity
    m_rsw = 0  # regulatory sweating
    q_res = 0  # heat loss due to respiration
    et = 0  # effective temperature
    e_req = 0  # evaporative heat loss required for tmp regulation
    r_ea = 0
    r_ecl = 0
    c_res = 0  # convective heat loss respiration

    pressure_in_atmospheres = p_atm / 101325
    length_time_simulation = 60  # length time simulation
    n_simulation = 1

    r_clo = 0.155 * clo  # thermal resistance of clothing, C M^2 /W
    f_a_cl = 1.0 + 0.15 * clo  # increase in body surface area due to clothing
    lr = 2.2 / pressure_in_atmospheres  # Lewis ratio
    rm = (met - wme) * met_factor  # metabolic rate
    m = met * met_factor  # metabolic rate

    e_comfort = 0.42 * (rm - met_factor)  # evaporative heat loss during comfort
    e_comfort = max(e_comfort, 0)

    i_cl = 1.0  # permeation efficiency of water vapour naked skin
    if clo > 0:
        i_cl = 0.45  # permeation efficiency of water vapour through the clothing layer

    if not w_max:  # if the user did not pass a value of w_max
        w_max = 0.38 * pow(air_speed, -0.29)  # critical skin wettedness naked
        if clo > 0:
            w_max = 0.59 * pow(air_speed, -0.08)  # critical skin wettedness clothed

    # h_cc corrected convective heat transfer coefficient
    h_cc = 3.0 * pow(pressure_in_atmospheres, 0.53)
    # h_fc forced convective heat transfer coefficient, W/(m2 °C)
    h_fc = 8.600001 * pow((air_speed * pressure_in_atmospheres), 0.53)
    h_cc = max(h_cc, h_fc)
    if not calculate_ce and met > 0.85:
        h_c_met = 5.66 * (met - 0.85) ** 0.39
        h_cc = max(h_cc, h_c_met)

    h_r = 4.7  # linearized radiative heat transfer coefficient
    h_t = h_r + h_cc  # sum of convective and radiant heat transfer coefficient W/(m2*K)
    r_a = 1.0 / (f_a_cl * h_t)  # resistance of air layer to dry heat
    t_op = (h_r * tr + h_cc * tdb) / h_t  # operative temperature

    t_body = alfa * t_skin + (1 - alfa) * t_core  # mean body temperature, °C

    # respiration
    q_res = 0.0023 * m * (44.0 - vapor_pressure)  # latent heat loss due to respiration
    c_res = 0.0014 * m * (34.0 - tdb)  # sensible convective heat loss respiration

    while n_simulation < length_time_simulation:
        n_simulation += 1

        iteration_limit = 150  # for following while loop
        # t_cl temperature of the outer surface of clothing
        t_cl = (r_a * t_skin + r_clo * t_op) / (r_a + r_clo)  # initial guess
        n_iterations = 0
        tc_converged = False

        while not tc_converged:
            # 0.95 is the clothing emissivity from ASHRAE fundamentals Ch. 9.7 Eq. 35
            if position == Postures.sitting.value:
                # 0.7 ratio between radiation area of the body and the body area
                h_r = 4.0 * 0.95 * sbc * ((t_cl + tr) / 2.0 + 273.15) ** 3.0 * 0.7
            else:  # if standing
                # 0.73 ratio between radiation area of the body and the body area
                h_r = 4.0 * 0.95 * sbc * ((t_cl + tr) / 2.0 + 273.15) ** 3.0 * 0.73
            h_t = h_r + h_cc
            r_a = 1.0 / (f_a_cl * h_t)
            t_op = (h_r * tr + h_cc * tdb) / h_t
            t_cl_new = (r_a * t_skin + r_clo * t_op) / (r_a + r_clo)
            if abs(t_cl_new - t_cl) <= 0.01:
                tc_converged = True
            t_cl = t_cl_new
            n_iterations += 1

            if n_iterations > iteration_limit:
                max_iteration_exceeded = "Max iterations exceeded"
                raise StopIteration(max_iteration_exceeded)

        q_sensible = (t_skin - t_op) / (r_a + r_clo)  # total sensible heat loss, W
        # hf_cs rate of energy transport between core and skin, W
        # 5.28 is the average body tissue conductance in W/(m2 C)
        # 1.163 is the thermal capacity of blood in Wh/(L C)
        hf_cs = (t_core - t_skin) * (5.28 + 1.163 * m_bl)
        s_core = m - hf_cs - q_res - c_res - wme  # rate of energy storage in the core
        s_skin = hf_cs - q_sensible - e_skin  # rate of energy storage in the skin
        tc_sk = 0.97 * alfa * body_weight  # thermal capacity skin
        tc_cr = 0.97 * (1 - alfa) * body_weight  # thermal capacity core
        d_t_sk = (s_skin * body_surface_area) / (
            tc_sk * 60.0
        )  # rate of change skin temperature °C per minute
        d_t_cr = (
            s_core * body_surface_area / (tc_cr * 60.0)
        )  # rate of change core temperature °C per minute
        t_skin = t_skin + d_t_sk
        t_core = t_core + d_t_cr
        t_body = alfa * t_skin + (1 - alfa) * t_core
        # sk_sig thermoregulatory control signal from the skin
        sk_sig = t_skin - temp_skin_neutral
        warm_sk = (sk_sig > 0) * sk_sig  # vasodilation signal
        colds = ((-1.0 * sk_sig) > 0) * (-1.0 * sk_sig)  # vasoconstriction signal
        # c_reg_sig thermoregulatory control signal from the skin, °C
        c_reg_sig = t_core - temp_core_neutral
        # c_warm vasodilation signal
        c_warm = (c_reg_sig > 0) * c_reg_sig
        # c_cold vasoconstriction signal
        c_cold = ((-1.0 * c_reg_sig) > 0) * (-1.0 * c_reg_sig)
        # bd_sig thermoregulatory control signal from the body
        bd_sig = t_body - temp_body_neutral
        warm_b = (bd_sig > 0) * bd_sig
        m_bl = (skin_blood_flow_neutral + c_dil * c_warm) / (1 + c_str * colds)
        m_bl = min(m_bl, max_skin_blood_flow)
        m_bl = max(m_bl, 0.5)
        m_rsw = c_sw * warm_b * math.exp(warm_sk / 10.7)  # regulatory sweating
        m_rsw = min(m_rsw, max_sweating)
        e_rsw = 0.68 * m_rsw  # heat lost by vaporization sweat
        r_ea = 1.0 / (lr * f_a_cl * h_cc)  # evaporative resistance air layer
        r_ecl = r_clo / (lr * i_cl)
        e_req = (
            rm - q_res - c_res - q_sensible
        )  # evaporative heat loss required for tmp regulation
        e_max = (math.exp(18.6686 - 4030.183 / (t_skin + 235.0)) - vapor_pressure) / (
            r_ea + r_ecl
        )
        if e_max == 0:  # added this otherwise e_rsw / e_max cannot be calculated
            e_max = 0.001
        p_rsw = e_rsw / e_max  # ratio heat loss sweating to max heat loss sweating
        w = 0.06 + 0.94 * p_rsw  # skin wetness
        e_diff = w * e_max - e_rsw  # vapor diffusion through skin
        if w > w_max:
            w = w_max
            p_rsw = w_max / 0.94
            e_rsw = p_rsw * e_max
            e_diff = 0.06 * (1.0 - p_rsw) * e_max
        if e_max < 0:
            e_diff = 0
            e_rsw = 0
            w = w_max
        e_skin = (
            e_rsw + e_diff
        )  # total evaporative heat loss sweating and vapor diffusion
        m_rsw = (
            e_rsw / 0.68
        )  # back calculating the mass of regulatory sweating as a function of e_rsw
        met_shivering = 19.4 * colds * c_cold  # met shivering W/m2
        m = rm + met_shivering
        alfa = 0.0417737 + 0.7451833 / (m_bl + 0.585417)

    q_skin = q_sensible + e_skin  # total heat loss from skin, W
    # p_s_sk saturation vapour pressure of water of the skin
    p_s_sk = math.exp(18.6686 - 4030.183 / (t_skin + 235.0))

    # standard environment - where _s at end of the variable names stands for standard
    h_r_s = h_r  # standard environment radiative heat transfer coefficient

    h_c_s = 3.0 * pow(pressure_in_atmospheres, 0.53)
    if not calculate_ce and met > 0.85:
        h_c_met = 5.66 * (met - 0.85) ** 0.39
        h_c_s = max(h_c_s, h_c_met)
    h_c_s = max(h_c_s, 3.0)

    h_t_s = (
        h_c_s + h_r_s
    )  # sum of convective and radiant heat transfer coefficient W/(m2*K)
    r_clo_s = (
        1.52 / ((met - wme / met_factor) + 0.6944) - 0.1835
    )  # thermal resistance of clothing, °C M^2 /W
    r_cl_s = 0.155 * r_clo_s  # thermal insulation of the clothing in M2K/W
    f_a_cl_s = 1.0 + k_clo * r_clo_s  # increase in body surface area due to clothing
    f_cl_s = 1.0 / (
        1.0 + 0.155 * f_a_cl_s * h_t_s * r_clo_s
    )  # ratio of surface clothed body over nude body
    i_m_s = 0.45  # permeation efficiency of water vapour through the clothing layer
    i_cl_s = (
        i_m_s * h_c_s / h_t_s * (1 - f_cl_s) / (h_c_s / h_t_s - f_cl_s * i_m_s)
    )  # clothing vapor permeation efficiency
    r_a_s = 1.0 / (f_a_cl_s * h_t_s)  # resistance of air layer to dry heat
    r_ea_s = 1.0 / (lr * f_a_cl_s * h_c_s)
    r_ecl_s = r_cl_s / (lr * i_cl_s)
    h_d_s = 1.0 / (r_a_s + r_cl_s)
    h_e_s = 1.0 / (r_ea_s + r_ecl_s)

    # calculate Standard Effective Temperature (SET)
    delta = 0.0001
    dx = 100.0
    set_old = round(t_skin - q_skin / h_d_s, 2)
    while abs(dx) > 0.01:
        err_1 = (
            q_skin
            - h_d_s * (t_skin - set_old)
            - w
            * h_e_s
            * (p_s_sk - 0.5 * (math.exp(18.6686 - 4030.183 / (set_old + 235.0))))
        )
        err_2 = (
            q_skin
            - h_d_s * (t_skin - (set_old + delta))
            - w
            * h_e_s
            * (
                p_s_sk
                - 0.5 * (math.exp(18.6686 - 4030.183 / (set_old + delta + 235.0)))
            )
        )
        _set = set_old - delta * err_1 / (err_2 - err_1)
        dx = _set - set_old
        set_old = _set

    # calculate Effective Temperature (ET)
    h_d = 1 / (r_a + r_clo)
    h_e = 1 / (r_ea + r_ecl)
    et_old = t_skin - q_skin / h_d
    delta = 0.0001
    dx = 100.0
    while abs(dx) > 0.01:
        err_1 = (
            q_skin
            - h_d * (t_skin - et_old)
            - w
            * h_e
            * (p_s_sk - 0.5 * (math.exp(18.6686 - 4030.183 / (et_old + 235.0))))
        )
        err_2 = (
            q_skin
            - h_d * (t_skin - (et_old + delta))
            - w
            * h_e
            * (p_s_sk - 0.5 * (math.exp(18.6686 - 4030.183 / (et_old + delta + 235.0))))
        )
        et = et_old - delta * err_1 / (err_2 - err_1)
        dx = et - et_old
        et_old = et

    tbm_l = (
        0.194 / met_to_w_m2
    ) * rm + 36.301  # lower limit for evaporative regulation
    tbm_h = (
        0.347 / met_to_w_m2
    ) * rm + 36.669  # upper limit for evaporative regulation

    t_sens = 0.4685 * (t_body - tbm_l)  # predicted thermal sensation
    if (t_body >= tbm_l) & (t_body < tbm_h):
        t_sens = w_max * 4.7 * (t_body - tbm_l) / (tbm_h - tbm_l)
    elif t_body >= tbm_h:
        t_sens = w_max * 4.7 + 0.4685 * (t_body - tbm_h)

    disc = (
        4.7 * (e_rsw - e_comfort) / (e_max * w_max - e_comfort - e_diff)
    )  # predicted thermal discomfort
    if disc <= 0:
        disc = t_sens

    # PMV Gagge
    pmv_gagge = (0.303 * math.exp(-0.036 * m) + 0.028) * (e_req - e_comfort - e_diff)

    # PMV SET
    dry_set = h_d_s * (t_skin - _set)
    e_req_set = rm - c_res - q_res - dry_set
    pmv_set = (0.303 * math.exp(-0.036 * m) + 0.028) * (e_req_set - e_comfort - e_diff)

    # Predicted  Percent  Satisfied  With  the  Level  of  Air  Movement
    # ps = 100 * (1.13 * (t_op**0.5) - 0.24 * t_op + 2.7 * (v**0.5) - 0.99 * v)

    return (
        _set,
        e_skin,
        e_rsw,
        e_max,
        q_sensible,
        q_skin,
        q_res,
        t_core,
        t_skin,
        m_bl,
        m_rsw,
        w,
        w_max,
        et,
        pmv_gagge,
        pmv_set,
        disc,
        t_sens,
    )


@vectorize(
    [  # signature: output(float64)
        float64(
            float64,
            float64,
            float64,
            float64,
            float64,
            float64,
            float64,
            float64,
            float64,
            float64,
        ),
    ],
    cache=True,
)
def _gagge_two_nodes_optimized_return_set(
    tdb,
    tr,
    v,
    met,
    clo,
    vapor_pressure,
    wme,
    body_surface_area,
    p_atm,
    position,
):
    return _gagge_two_nodes_optimized(
        tdb=tdb,
        tr=tr,
        v=v,
        met=met,
        clo=clo,
        vapor_pressure=vapor_pressure,
        wme=wme,
        body_surface_area=body_surface_area,
        p_atm=p_atm,
        position=position,
        calculate_ce=True,
    )[0]



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import AnkleDraftInputs
from pythermalcomfort.classes_return import AnkleDraft
from pythermalcomfort.models.pmv_ppd_ashrae import pmv_ppd_ashrae
from pythermalcomfort.utilities import (
    Models,
    Units,
    _check_standard_compliance_array,
    units_converter,
)


def ankle_draft(
    tdb: float | list[float],
    tr: float | list[float],
    vr: float | list[float],
    rh: float | list[float],
    met: float | list[float],
    clo: float | list[float],
    v_ankle: float | list[float],
    units: str = Units.SI.value,
) -> AnkleDraft:
    """Calculate the percentage of thermally dissatisfied people with the
    ankle draft (0.1 m) above floor level [Liu2017]_.

    This equation is only applicable for vr < 0.2 m/s (40 fps).

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, default in [°C] or [°F] if `units` = 'IP'.

        .. note::
            The air temperature is the average value over two heights: 0.6 m (24 in.)
            and 1.1 m (43 in.) for seated occupants, and 1.1 m (43 in.) and 1.7 m (67 in.) for standing occupants.

    tr : float or list of floats
        Mean radiant temperature, default in [°C] or [°F] if `units` = 'IP'.

    vr : float or list of floats
        Relative air speed, default in [m/s] or [fps] if `units` = 'IP'.

        .. note::
            `vr` is the relative air speed caused by body movement and not the air speed measured by the air speed sensor.
            The relative air speed is the sum of the average air speed measured by the sensor plus the activity-generated air speed (Vag).
            Vag is the activity-generated air speed caused by motion of individual body parts.
            `vr` can be calculated using the function :py:meth:`pythermalcomfort.utilities.v_relative`.

    rh : float or list of floats
        Relative humidity, [%].

    met : float or list of floats
        Metabolic rate, [met].

    clo : float or list of floats
        Clothing insulation, [clo].

        .. note::
            this is the basic insulation also known as the intrinsic clothing insulation value of the
            clothing ensemble (`I`:sub:`cl,r`), this is the thermal insulation from the skin
            surface to the outer clothing surface, including enclosed air layers, under actual
            environmental conditions. This value is not the total insulation (`I`:sub:`T,r`).
            The dynamic clothing insulation, clo, can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.clo_dynamic_ashrae`.

    v_ankle : float or list of floats
        Air speed at 0.1 m (4 in.) above the floor, default in [m/s] or [fps] if `units` = 'IP'.

    units : {'SI', 'IP'}
        Select the SI (International System of Units) or the IP (Imperial Units) system.

    Returns
    -------
    AnkleDraft
        Dataclass containing the results of the ankle draft calculation. See :py:class:`~pythermalcomfort.classes_return.AnkleDraft` for more details.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import ankle_draft

        results = ankle_draft(25, 25, 0.2, 50, 1.2, 0.5, 0.3, units="SI")
        print(results)
        # AnkleDraft(ppd_ad=18.5, acceptability=True)

    """
    # Validate inputs using the AnkleDraftInputs class
    AnkleDraftInputs(
        tdb=tdb,
        tr=tr,
        vr=vr,
        rh=rh,
        met=met,
        clo=clo,
        v_ankle=v_ankle,
        units=units,
    )

    # Convert lists to numpy arrays
    tdb = np.array(tdb)
    tr = np.array(tr)
    vr = np.array(vr)
    rh = np.array(rh)
    met = np.array(met)
    clo = np.array(clo)
    v_ankle = np.array(v_ankle)

    if units.upper() == Units.IP.value:
        tdb, tr, vr, v_ankle = units_converter(tdb=tdb, tr=tr, vr=vr, vel=v_ankle)

    tdb_valid, tr_valid, v_valid, v_limited = _check_standard_compliance_array(
        standard=Models.ashrae_55_2023.value,
        tdb=tdb,
        tr=tr,
        v_limited=vr,
        v=vr,
    )

    if np.all(np.isnan(v_limited)):
        raise ValueError(
            "This equation is only applicable for air speed lower than 0.2 m/s",
        )

    tsv = pmv_ppd_ashrae(
        tdb,
        tr,
        vr,
        rh,
        met,
        clo,
        model=Models.ashrae_55_2023.value,
    ).pmv
    ppd_val = np.around(
        np.exp(-2.58 + 3.05 * v_ankle - 1.06 * tsv)
        / (1 + np.exp(-2.58 + 3.05 * v_ankle - 1.06 * tsv))
        * 100,
        1,
    )
    acceptability = ppd_val <= 20
    return AnkleDraft(ppd_ad=ppd_val, acceptability=acceptability)



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import HumidexInputs, HumidexModels
from pythermalcomfort.classes_return import Humidex
from pythermalcomfort.utilities import dew_point_tmp


def humidex(
    tdb: float | list[float],
    rh: float | list[float],
    model: str = "rana",
    round_output: bool = True,
) -> Humidex:
    """Calculate the humidex (short for "humidity index"). It has been
    developed by the Canadian Meteorological service. It was introduced in 1965
    and then it was revised by Masterson and Richardson (1979) [Masterson1979]_. It aims
    to describe how hot, humid weather is felt by the average person. The
    Humidex differs from the heat index in being related to the dew point
    rather than relative humidity [Havenith2016]_.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    rh : float or list of floats
        Relative humidity, [%].
    model : str, optional
        The model to be used for the calculation. Options are 'rana' and 'masterson'. Defaults to 'rana'.

        .. note::
            The 'rana' model is the Humidex model proposed by `Rana et al. (2013)`_.
            The 'masterson' model is the Humidex model proposed by Masterson and Richardson (1979) [Masterson1979]_.

            .. _Rana et al. (2013): https://doi.org/10.1016/j.enbuild.2013.04.019
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    Humidex
        A dataclass containing the Humidex value and its discomfort category. See :py:class:`~pythermalcomfort.classes_return.Humidex` for more details.
        To access the `humidex` and `discomfort` values, use the respective attributes of the returned `Humidex` instance, e.g., `result.humidex`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import humidex

        result = humidex(tdb=25, rh=50)
        print(result.humidex)  # 28.2
        print(result.discomfort)  # Little or no discomfort

        result = humidex(tdb=[25, 30], rh=[50, 60], round_output=False)
        print(result.humidex)  # [28.2, 39.1]
        print(result.discomfort)
        # ['Little or no discomfort', 'Evident discomfort']

    """
    # Validate inputs using the HumidexInputs class
    HumidexInputs(
        tdb=tdb,
        rh=rh,
        round_output=round_output,
    )

    tdb = np.array(tdb)
    rh = np.array(rh)

    if np.any(rh > 100) or np.any(rh < 0):
        raise ValueError("Relative humidity must be between 0 and 100%")

    if model not in HumidexModels._value2member_map_:
        raise ValueError(
            "Invalid model. The model must be either 'rana' or 'masterson'",
        )

    hi = tdb + 5 / 9 * ((6.112 * 10 ** (7.5 * tdb / (237.7 + tdb)) * rh / 100) - 10)
    if model == HumidexModels.masterson.value:
        hi = tdb + 5 / 9 * (
            6.11
            * np.exp(
                5417.753 * (1 / 273.15 - 1 / (dew_point_tmp(tdb=tdb, rh=rh) + 273.15)),
            )
            - 10
        )

    if round_output:
        hi = np.around(hi, 1)

    stress_category = np.full_like(hi, "Heat stroke probable", dtype=object)
    stress_category[hi <= 30] = "Little or no discomfort"
    stress_category[(hi > 30) & (hi <= 35)] = "Noticeable discomfort"
    stress_category[(hi > 35) & (hi <= 40)] = "Evident discomfort"
    stress_category[(hi > 40) & (hi <= 45)] = "Intense discomfort; avoid exertion"
    stress_category[(hi > 45) & (hi <= 54)] = "Dangerous discomfort"

    return Humidex(humidex=hi, discomfort=stress_category)



import numpy as np
import pytest

from pythermalcomfort.models.wind_chill_temperature import wind_chill_temperature


class TestWct:
    """Test cases for the wind chill temperature (WCT) model."""

    # TODO this tests are not synced with comf R

    def test_wct_results(self) -> None:
        """Test that the function calculates WCT correctly for various inputs."""
        assert np.isclose(wind_chill_temperature(tdb=-20, v=5).wct, -24.3, atol=0.1)
        assert np.isclose(wind_chill_temperature(tdb=-20, v=15).wct, -29.1, atol=0.1)
        assert np.isclose(wind_chill_temperature(tdb=-20, v=60).wct, -36.5, atol=0.1)

    # Calculate WCT correctly for single float inputs of temperature and wind speed
    def test_wct_single_float_inputs(self) -> None:
        """Test that the function calculates WCT correctly for single float values of tdb and v."""
        # Test with single float values
        result = wind_chill_temperature(tdb=-5.0, v=5.5)

        # Expected value calculated using the formula:
        # 13.12 + 0.6215 * tdb - 11.37 * v**0.16 + 0.3965 * tdb * v**0.16
        expected = -7.5

        assert isinstance(result.wct, float)
        assert round(result.wct, 1) == expected

    # Handle empty lists for temperature and wind speed inputs
    def test_wct_empty_lists(self) -> None:
        """Test that the function handles empty lists for tdb and v inputs."""
        # Test with empty lists
        assert np.allclose(
            wind_chill_temperature(tdb=[], v=[]).wct, np.array([]), equal_nan=True
        )

    # Calculate WCT correctly for lists of temperature and wind speed values
    def test_wct_list_inputs(self) -> None:
        """Test that the function calculates WCT correctly for lists of tdb and v values."""
        # Test with list of values
        result = wind_chill_temperature(
            tdb=[-5.0, -10.0], v=[5.5, 10.0], round_output=True
        )

        # Expected values calculated using the formula:
        # 13.12 + 0.6215 * tdb - 11.37 * v**0.16 + 0.3965 * tdb * v**0.16
        expected = [-7.5, -15.3]

        assert isinstance(result.wct, np.ndarray)
        assert np.allclose(result.wct, expected, atol=0.1)

    # Handle non-numeric input values
    def test_wct_non_numeric_inputs(self) -> None:
        """Function raises a TypeError if non-numeric values are inputs."""
        # Test with non-numeric values for tdb and v
        with pytest.raises(TypeError):
            wind_chill_temperature(tdb="invalid", v=5.5)

        with pytest.raises(TypeError):
            wind_chill_temperature(tdb=-5.0, v="invalid")

        with pytest.raises(TypeError):
            wind_chill_temperature(tdb="invalid", v="invalid")



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import ESIInputs
from pythermalcomfort.classes_return import ESI


def esi(
    tdb: float | list[float],
    rh: float | list[float],
    sol_radiation_global: float | list[float],
    round_output: bool = True,
) -> ESI:
    """Calculate the Environmental Stress Index (ESI) [Moran2001]_.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    rh: float or list of floats
        Relative humidity, [%].
    sol_radiation_global: float or list of floats
        Global solar radiation, [W/m2].
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    ESI
        A dataclass containing the Environmental Stress Index. See :py:class:`~pythermalcomfort.classes_return.ESI` for more details.
        To access the `esi` value, use the `esi` attribute of the returned `ESI` instance, e.g., `result.esi`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import esi

        result = esi(tdb=30.2, rh=42.2, sol_radiation_global=766)
        print(result.esi)  # 26.2

        result = esi(tdb=[30.2, 27.0], rh=[42.2, 68.8], sol_radiation_global=[766, 289])
        print(result.esi)  # [26.2, 25.6]

    """
    ESIInputs(
        tdb=tdb,
        rh=rh,
        sol_radiation_global=sol_radiation_global,
        round_output=round_output,
    )

    tdb = np.array(tdb)
    rh = np.array(rh)
    sol_radiation_global = np.array(sol_radiation_global)

    _esi = (
        0.63 * tdb
        - 0.03 * rh
        + 0.002 * sol_radiation_global
        + 0.0054 * (tdb * rh)
        - 0.073 * (0.1 + sol_radiation_global) ** (-1)
    )

    if round_output:
        _esi = np.round(_esi, 1)

    return ESI(esi=_esi)



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import NETInputs
from pythermalcomfort.classes_return import NET


def net(
    tdb: float | list[float],
    rh: float | list[float],
    v: float | list[float],
    round_output: bool = True,
) -> NET:
    """Calculate the Normal Effective Temperature (NET). Missenard (1933)
    devised a formula for calculating effective temperature. The index
    establishes a link between the same condition of the organism's
    thermoregulatory capability (warm and cold perception) and the surrounding
    environment's temperature and humidity. The index is calculated as a
    function of three meteorological factors: air temperature, relative
    humidity of air, and wind speed. This index allows to calculate the
    effective temperature felt by a person. Missenard original equation was
    then used to calculate the Normal Effective Temperature (NET), by
    considering normal atmospheric pressure and a normal human body temperature
    (37°C). The NET is still in use in Germany, where medical check-ups for
    subjects working in the heat are decided on by prevailing levels of ET,
    depending on metabolic rates. The NET is also constantly monitored by the
    Hong Kong Observatory [Blazejczyk2012]_. In central Europe the following thresholds are
    in use: <1°C = very cold; 1-9 = cold; 9-17 = cool; 17-21 = fresh; 21-23 = comfortable;
    23-27 = warm; >27°C = hot [Blazejczyk2012]_.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    rh : float or list of floats
        Relative humidity, [%].
    v : float or list of floats
        Wind speed [m/s] at 1.2 m above the ground.
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    NET
        A dataclass containing the Normal Effective Temperature. See :py:class:`~pythermalcomfort.classes_return.Net` for more details.
        To access the `net` value, use the `net` attribute of the returned `Net` instance, e.g., `result.net`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import net

        result = net(tdb=37, rh=100, v=0.1)
        print(result.net)  # 37.0

        result = net(tdb=[37, 30], rh=[100, 60], v=[0.1, 0.5], round_output=False)
        print(result.net)  # [37.0, 26.38977535]

    """
    # Validate inputs using the NetInputs class
    NETInputs(
        tdb=tdb,
        rh=rh,
        v=v,
        round_output=round_output,
    )

    tdb = np.array(tdb)
    rh = np.array(rh)
    v = np.array(v)

    frac = 1.0 / (1.76 + 1.4 * v**0.75)
    et = 37 - (37 - tdb) / (0.68 - 0.0014 * rh + frac) - 0.29 * tdb * (1 - 0.01 * rh)

    if round_output:
        et = np.around(et, 1)

    return NET(net=et)



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import EPMVInputs
from pythermalcomfort.classes_return import EPMV
from pythermalcomfort.models.pmv_ppd_iso import pmv_ppd_iso
from pythermalcomfort.utilities import Models, Units


def pmv_e(
    tdb: float | list[float],
    tr: float | list[float],
    vr: float | list[float],
    rh: float | list[float],
    met: float | list[float],
    clo: float | list[float],
    e_coefficient: float | list[float],
    wme: float | list[float] = 0,
    units: str = Units.SI.value,
    limit_inputs: bool = True,
) -> EPMV:
    """Return Adjusted Predicted Mean Votes with Expectancy Factor (ePMV).
    This index was developed by Fanger, P. O. et al. (2002). In non-air-
    conditioned buildings in warm climates, occupants may sense the warmth as
    being less severe than the PMV predicts. The main reason is low
    expectations, but a metabolic rate that is estimated too high can also
    contribute to explaining the difference. An extension of the PMV model that
    includes an expectancy factor is introduced for use in non-air-conditioned
    buildings in warm climates [Fanger2002]_.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, default in [°C] in [°F] if `units` = 'IP'.
    tr : float or list of floats
        Mean radiant temperature, default in [°C] in [°F] if `units` = 'IP'.
    vr : float or list of floats
        Relative air speed, default in [m/s] in [fps] if `units` = 'IP'.

        .. note::
            vr is the relative air speed caused by body movement and not the air
            speed measured by the air speed sensor. The relative air speed is the sum of the
            average air speed measured by the sensor plus the activity-generated air speed
            (Vag). Where Vag is the activity-generated air speed caused by motion of
            individual body parts. vr can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.v_relative`.

    rh : float or list of floats
        Relative humidity, [%].
    met : float or list of floats
        Metabolic rate, [met].
    clo : float or list of floats
        Clothing insulation, [clo].

        .. note::
            this is the basic insulation also known as the intrinsic clothing insulation value of the
            clothing ensemble (`I`:sub:`cl,r`), this is the thermal insulation from the skin
            surface to the outer clothing surface, including enclosed air layers, under actual
            environmental conditions. This value is not the total insulation (`I`:sub:`T,r`).
            The dynamic clothing insulation, clo, can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.clo_dynamic_iso`.

    e_coefficient : float or list of floats
        Expectancy factor.
    wme : float or list of floats, optional
        External work, [met]. Defaults to 0.
    units : {'SI', 'IP'}, optional
        Select the SI (International System of Units) or the IP (Imperial Units) system.
        Supported values are 'SI' and 'IP'. Defaults to 'SI'.
    limit_inputs : bool, optional
        By default, if the inputs are outside the standard applicability limits the
        function returns nan. If False, returns pmv and ppd values even if input values are
        outside the applicability limits of the model. Defaults to True.

        .. note::
            The ISO 7730 2005 limits are 10 < tdb [°C] < 30, 10 < tr [°C] < 40,
            0 < vr [m/s] < 1, 0.8 < met [met] < 4, 0 < clo [clo] < 2, and -2 < PMV < 2.

    Returns
    -------
    EPMV
        A dataclass containing the Adjusted Predicted Mean Votes with Expectancy Factor. See :py:class:`~pythermalcomfort.classes_return.EPMV` for more details.
        To access the `e_pmv` value, use the `e_pmv` attribute of the returned `e_pmv` instance, e.g., `result.e_pmv`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import pmv_e
        from pythermalcomfort.utilities import v_relative, clo_dynamic_iso

        tdb = 28
        tr = 28
        rh = 50
        v = 0.1
        met = 1.4
        clo = 0.5
        # calculate relative air speed
        v_r = v_relative(v=v, met=met)
        # Calculate dynamic clothing
        clo_d = clo_dynamic_iso(clo=clo, met=met, v=v)
        results = pmv_e(tdb, tr, v_r, rh, met, clo_d, e_coefficient=0.6)
        print(results.e_pmv)  # 0.48

    """
    # Validate inputs using the EPMVInputs class
    EPMVInputs(
        tdb=tdb,
        tr=tr,
        vr=vr,
        rh=rh,
        met=met,
        clo=clo,
        e_coefficient=e_coefficient,
        wme=wme,
        units=units,
    )

    default_kwargs = {"units": units, "limit_inputs": limit_inputs}
    met = np.array(met)
    _pmv = pmv_ppd_iso(
        tdb=tdb,
        tr=tr,
        vr=vr,
        rh=rh,
        met=met,
        clo=clo,
        wme=wme,
        model=Models.iso_7730_2005.value,
        **default_kwargs,
    ).pmv
    met = np.where(_pmv > 0, met * (1 + _pmv * (-0.067)), met)
    _pmv = pmv_ppd_iso(
        tdb=tdb,
        tr=tr,
        vr=vr,
        rh=rh,
        met=met,
        clo=clo,
        wme=wme,
        model=Models.iso_7730_2005.value,
        **default_kwargs,
    ).pmv

    e_pmv_value = np.around(_pmv * e_coefficient, 2)

    return EPMV(e_pmv=e_pmv_value)



from pythermalcomfort.models.adaptive_ashrae import adaptive_ashrae
from pythermalcomfort.models.adaptive_en import adaptive_en
from pythermalcomfort.models.ankle_draft import ankle_draft
from pythermalcomfort.models.at import at
from pythermalcomfort.models.clo_tout import clo_tout
from pythermalcomfort.models.cooling_effect import cooling_effect
from pythermalcomfort.models.discomfort_index import discomfort_index
from pythermalcomfort.models.esi import esi
from pythermalcomfort.models.heat_index_lu import heat_index_lu
from pythermalcomfort.models.heat_index_rothfusz import heat_index_rothfusz
from pythermalcomfort.models.humidex import humidex
from pythermalcomfort.models.jos3 import JOS3
from pythermalcomfort.models.net import net
from pythermalcomfort.models.pet_steady import pet_steady
from pythermalcomfort.models.phs import phs
from pythermalcomfort.models.pmv_a import pmv_a
from pythermalcomfort.models.pmv_athb import pmv_athb
from pythermalcomfort.models.pmv_e import pmv_e
from pythermalcomfort.models.pmv_ppd_ashrae import pmv_ppd_ashrae
from pythermalcomfort.models.pmv_ppd_iso import pmv_ppd_iso
from pythermalcomfort.models.set_tmp import set_tmp
from pythermalcomfort.models.solar_gain import solar_gain
from pythermalcomfort.models.thi import thi
from pythermalcomfort.models.two_nodes_gagge import two_nodes_gagge
from pythermalcomfort.models.two_nodes_gagge_ji import two_nodes_gagge_ji
from pythermalcomfort.models.two_nodes_gagge_sleep import two_nodes_gagge_sleep
from pythermalcomfort.models.use_fans_heatwaves import use_fans_heatwaves
from pythermalcomfort.models.utci import utci
from pythermalcomfort.models.vertical_tmp_grad_ppd import vertical_tmp_grad_ppd
from pythermalcomfort.models.wbgt import wbgt
from pythermalcomfort.models.wci import wci
from pythermalcomfort.models.wind_chill_temperature import wind_chill_temperature
from pythermalcomfort.models.work_capacity_dunne import work_capacity_dunne
from pythermalcomfort.models.work_capacity_hothaps import work_capacity_hothaps
from pythermalcomfort.models.work_capacity_iso import work_capacity_iso
from pythermalcomfort.models.work_capacity_niosh import work_capacity_niosh

__all__ = [
    "JOS3",
    "adaptive_ashrae",
    "adaptive_en",
    "ankle_draft",
    "at",
    "clo_tout",
    "cooling_effect",
    "discomfort_index",
    "esi",
    "heat_index_lu",
    "heat_index_rothfusz",
    "humidex",
    "net",
    "pet_steady",
    "phs",
    "pmv_a",
    "pmv_athb",
    "pmv_e",
    "pmv_ppd_ashrae",
    "pmv_ppd_iso",
    "set_tmp",
    "solar_gain",
    "thi",
    "two_nodes_gagge",
    "two_nodes_gagge_ji",
    "two_nodes_gagge_sleep",
    "use_fans_heatwaves",
    "utci",
    "vertical_tmp_grad_ppd",
    "wbgt",
    "wci",
    "wind_chill_temperature",
    "work_capacity_dunne",
    "work_capacity_hothaps",
    "work_capacity_iso",
    "work_capacity_niosh",
]



from pprint import pprint

from pythermalcomfort.models import adaptive_en
from pythermalcomfort.utilities import running_mean_outdoor_temperature

result = adaptive_en(tdb=25, tr=25, t_running_mean=24, v=0.1)

pprint(result)

result = adaptive_en(tdb=22.5, tr=22.5, t_running_mean=24, v=0.1)

pprint(result)

comf_tmp = result["tmp_cmf"]

result = adaptive_en(tdb=72.5, tr=72.5, t_running_mean=75, v=0.1, units="IP")
pprint(result)

rmt_value = running_mean_outdoor_temperature([29, 28, 30, 29, 28, 30, 27], alpha=0.9)

result = adaptive_en(tdb=25, tr=25, t_running_mean=rmt_value, v=0.3)
pprint(result)



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import PMVPPDInputs
from pythermalcomfort.classes_return import PMVPPD
from pythermalcomfort.models._pmv_ppd_optimized import _pmv_ppd_optimized
from pythermalcomfort.models.cooling_effect import cooling_effect
from pythermalcomfort.shared_functions import mapping
from pythermalcomfort.utilities import (
    Models,
    Units,
    _check_standard_compliance_array,
    units_converter,
)


def pmv_ppd_ashrae(
    tdb: float | list[float],
    tr: float | list[float],
    vr: float | list[float],
    rh: float | list[float],
    met: float | list[float],
    clo: float | list[float],
    wme: float | list[float] = 0,
    model: str = Models.ashrae_55_2023.value,
    units: str = Units.SI.value,
    limit_inputs: bool = True,
    airspeed_control: bool = True,
    round_output: bool = True,
) -> PMVPPD:
    """Return Predicted Mean Vote (PMV) and Predicted Percentage of Dissatisfied (PPD)
    calculated in accordance with the ASHRAE 55 Standard.

    While the PMV equation is the same for both the ISO and ASHRAE standards, in the
    ASHRAE 55 PMV equation, the SET is used to calculate the cooling effect first,
    this is then subtracted from both the air and mean radiant temperatures, and the
    differences are used as input to the PMV model, while the airspeed is set to 0.1m/s.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C] in [°F] if `units` = 'IP'
    tr : float or list of floats
        Mean radiant temperature, [°C] in [°F] if `units` = 'IP'
    vr : float or list of floats
        Relative air speed, [m/s] in [fps] if `units` = 'IP'

        .. note::
            vr is the relative air speed caused by body movement and not the air
            speed measured by the air speed sensor. The relative air speed is the sum of the
            average air speed measured by the sensor plus the activity-generated air speed
            (Vag). Where Vag is the activity-generated air speed caused by motion of
            individual body parts. vr can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.v_relative`.

    rh : float or list of floats
        Relative humidity, [%].
    met : float or list of floats
        Metabolic rate, [met].
    clo : float or list of floats
        Clothing insulation, [clo].

        .. note::
            this is the basic insulation also known as the intrinsic clothing insulation value of the
            clothing ensemble (`I`:sub:`cl,r`), this is the thermal insulation from the skin
            surface to the outer clothing surface, including enclosed air layers, under actual
            environmental conditions. This value is not the total insulation (`I`:sub:`T,r`).
            The dynamic clothing insulation, clo, can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.clo_dynamic_ashrae`.

    wme : float or list of floats, optional
        External work, [met]. Defaults to 0.
    model : str, optional
        Select the version of the ASHRAE 55 Standard to use. Currently, the only option available is "55-2023".
    units : str, optional
        Select the SI (International System of Units) or the IP (Imperial Units) system. Supported values are
        'SI' and 'IP'. Defaults to 'SI'.
    limit_inputs : bool, optional
        If True, limits the inputs to the standard applicability limits. Defaults to True.

        .. note::
            By default, if the inputs are outside the standard applicability limits the
            function returns nan. If False returns pmv and ppd values even if input values are
            outside the applicability limits of the model.

            The ASHRAE 55 2020 limits are 10 < tdb [°C] < 40, 10 < tr [°C] < 40,
            0 < vr [m/s] < 2, 1 < met [met] < 4, and 0 < clo [clo] < 1.5.

    airspeed_control : bool, optional
        By default, this function assumes that the
        occupant has control over the airspeed. In this case, the ASHRAE 55 Standard does
        not impose any airspeed limits. On the other hand, if the occupant has no control
        over the airspeed, the ASHRAE 55 imposes an upper limit for v which varies as a
        function of the operative temperature, for more information please consult the
        Standard.
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    PMVPPD
        A dataclass containing the Predicted Mean Vote and Predicted Percentage of
        Dissatisfied. See :py:class:`~pythermalcomfort.classes_return.PMVPPD` for
        more details. To access the `pmv`, `ppd`, `tsv` values, use the corresponding
        attributes of the returned `PMVPPD` instance, e.g., `result.pmv`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import pmv_ppd_ashrae
        from pythermalcomfort.utilities import v_relative, clo_dynamic_ashrae

        tdb = 25
        tr = 25
        rh = 50
        v = 0.1
        met = 1.4
        clo = 0.5
        # calculate relative air speed
        v_r = v_relative(v=v, met=met)
        # calculate dynamic clothing
        clo_d = clo_dynamic_ashrae(clo=clo, met=met)
        results = pmv_ppd_ashrae(
            tdb=tdb, tr=tr, vr=v_r, rh=rh, met=met, clo=clo_d, model="55-2023"
        )
        print(results.pmv)  # 0.0
        print(results.ppd)  # 5.0

        result = pmv_ppd_ashrae(
            tdb=[22, 25],
            tr=25,
            vr=0.1,
            rh=50,
            met=1.4,
            clo=0.5,
            model="55-2023",
        )
        print(result.pmv)  # [-0.  0.41]
        print(result.ppd)  # [5.  8.5]

    """
    # Validate inputs using the PMVPPDInputs class
    PMVPPDInputs(
        tdb=tdb,
        tr=tr,
        vr=vr,
        rh=rh,
        met=met,
        clo=clo,
        wme=wme,
        units=units,
        limit_inputs=limit_inputs,
        airspeed_control=airspeed_control,
    )

    tdb = np.array(tdb)
    tr = np.array(tr)
    rh = np.array(rh)
    vr = np.array(vr)
    met = np.array(met)
    clo = np.array(clo)
    wme = np.array(wme)

    if units.upper() == Units.IP.value:
        tdb, tr, vr = units_converter(tdb=tdb, tr=tr, v=vr)

    model = model.lower()
    if model not in [Models.ashrae_55_2023.value]:
        error_msg = (
            f"Model '{model}' is not supported. "
            f"Supported models are: {Models.ashrae_55_2023.value}."
        )
        raise ValueError(error_msg)

    (
        tdb_valid,
        tr_valid,
        v_valid,
        met_valid,
        clo_valid,
    ) = _check_standard_compliance_array(
        standard=Models.ashrae_55_2023.value,
        tdb=tdb,
        tr=tr,
        v=vr,
        met=met,
        clo=clo,
        airspeed_control=airspeed_control,
    )

    # if v_r is higher than 0.1 follow methodology ASHRAE Appendix H, H3
    ce = np.where(
        vr > 0.1,
        cooling_effect(tdb=tdb, tr=tr, vr=vr, rh=rh, met=met, clo=clo, wme=wme).ce,
        0.0,
    )

    tdb = tdb - ce
    tr = tr - ce
    vr = np.where(ce > 0, 0.1, vr)

    pmv_array = _pmv_ppd_optimized(tdb, tr, vr, rh, met, clo, wme)

    ppd_array = 100.0 - 95.0 * np.exp(
        -0.03353 * pmv_array**4.0 - 0.2179 * pmv_array**2.0,
    )

    # Checks that inputs are within the bounds accepted by the model if not return nan
    if limit_inputs:
        all_valid = ~(
            np.isnan(tdb_valid)
            | np.isnan(tr_valid)
            | np.isnan(v_valid)
            | np.isnan(met_valid)
            | np.isnan(clo_valid)
        )
        pmv_array = np.where(all_valid, pmv_array, np.nan)
        ppd_array = np.where(all_valid, ppd_array, np.nan)

    if round_output:
        pmv_array = np.round(pmv_array, 2)
        ppd_array = np.round(ppd_array, 1)

    thermal_sensation = {
        -2.5: "Cold",
        -1.5: "Cool",
        -0.5: "Slightly Cool",
        0.5: "Neutral",
        1.5: "Slightly Warm",
        2.5: "Warm",
        10: "Hot",
    }

    return PMVPPD(
        pmv=pmv_array,
        ppd=ppd_array,
        tsv=mapping(pmv_array, thermal_sensation, right=False),
    )



"""In this example I am calculating the Standard Effective Temperature (SET) with the pythermalcomfort package."""

from pythermalcomfort.models import set_tmp

set_value = set_tmp(tdb=25, tr=25, v=0.3, rh=60, met=1.2, clo=0.5)
print(f"set_value= {set_value.set}")

set_value = set_tmp(tdb=25, tr=25, v=0.5, rh=60, met=1.2, clo=0.5)
print(f"set_value= {set_value.set}")



from __future__ import annotations

from typing import Literal

import numpy as np

from pythermalcomfort.classes_input import APMVInputs
from pythermalcomfort.classes_return import APMV
from pythermalcomfort.models.pmv_ppd_iso import pmv_ppd_iso
from pythermalcomfort.utilities import Models, Units


def pmv_a(
    tdb: float | list[float],
    tr: float | list[float],
    vr: float | list[float],
    rh: float | list[float],
    met: float | list[float],
    clo: float | list[float],
    a_coefficient: float,
    wme: float | list[float] = 0,
    units: Literal["SI", "IP"] = Units.SI.value,
    limit_inputs: bool = True,
) -> APMV:
    """Return Adaptive Predicted Mean Vote (aPMV) [Yao2009]_. This index was
    developed by Yao, R. et al. (2009). The model takes into account factors
    such as culture, climate, social, psychological, and behavioral
    adaptations, which have an impact on the senses used to detect thermal
    comfort. This model uses an adaptive coefficient (λ) representing the
    adaptive factors that affect the sense of thermal comfort.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, default in [°C] or [°F] if `units` = 'IP'.
    tr : float or list of floats
        Mean radiant temperature, default in [°C] or [°F] if `units` = 'IP'.
    vr : float or list of floats
        Relative air speed, default in [m/s] or [fps] if `units` = 'IP'.

        .. note::
            vr is the sum of the average air speed measured by the sensor and the activity-generated air speed (Vag).
            Calculate vr using :py:meth:`pythermalcomfort.utilities.v_relative`.

    rh : float or list of floats
        Relative humidity, [%].
    met : float or list of floats
        Metabolic rate, [met].
    clo : float or list of floats
        Clothing insulation, [clo].

        .. note::
            Correct for body movement effects using :py:meth:`pythermalcomfort.utilities.clo_dynamic_iso`.

    a_coefficient : float
        Adaptive coefficient.
    wme : float or list of floats, optional
        External work, [met], default is 0.
    units : str, optional
        Units system, 'SI' or 'IP'. Defaults to 'SI'.
    limit_inputs : bool, optional
        If True, returns nan for inputs outside standard limits. Defaults to True.

        .. note::
            ISO 7730 2005 limits: 10 < tdb [°C] < 30, 10 < tr [°C] < 40, 0 < vr [m/s] < 1, 0.8 < met [met] < 4,
            0 < clo [clo] < 2, -2 < PMV < 2.

    Returns
    -------
    APMV
        A dataclass containing the Predicted Mean Vote (a_pmv). See
        :py:class:`~pythermalcomfort.classes_return.AdaptivePMV` for more details.
        To access the `a_pmv` value, use the `a_pmv` attribute of the returned `AdaptivePMV` instance,
        e.g., `result.a_pmv`.

    Examples
    --------
    .. code-block:: python
        :emphasize-lines: 9,12,14

        from pythermalcomfort.models import pmv_a
        from pythermalcomfort.utilities import v_relative, clo_dynamic_iso

        v = 0.1
        met = 1.4
        clo = 0.5

        # Calculate relative air speed
        v_r = v_relative(v=v, met=met)

        # Calculate dynamic clothing
        clo_d = clo_dynamic_iso(clo=clo, met=met, v=v)

        results = pmv_a(
            tdb=28,
            tr=28,
            vr=v_r,
            rh=50,
            met=met,
            clo=clo_d,
            a_coefficient=0.293,
        )
        print(results)  # AdaptivePMV(a_pmv=0.74)
        print(results.a_pmv)  # 0.71

    """
    # Validate inputs using the APMVInputs class
    APMVInputs(
        tdb=tdb,
        tr=tr,
        vr=vr,
        rh=rh,
        met=met,
        clo=clo,
        a_coefficient=a_coefficient,
        wme=wme,
        units=units,
    )

    _pmv = pmv_ppd_iso(
        tdb,
        tr,
        vr,
        rh,
        met,
        clo,
        wme,
        model=Models.iso_7730_2005.value,
        units=units,
        limit_inputs=limit_inputs,
    ).pmv

    pmv_value = np.around(_pmv / (1 + a_coefficient * _pmv), 2)

    return APMV(a_pmv=pmv_value)



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import ATHBInputs
from pythermalcomfort.classes_return import ATHB
from pythermalcomfort.models._pmv_ppd_optimized import _pmv_ppd_optimized
from pythermalcomfort.utilities import met_to_w_m2


def pmv_athb(
    tdb: float | list[float],
    tr: float | list[float],
    vr: float | list[float],
    rh: float | list[float],
    met: float | list[float],
    t_running_mean: float | list[float],
    clo: bool | float | list[float] = False,
) -> ATHB:
    """Return the PMV value calculated with the Adaptive Thermal Heat Balance
    Framework [Schweiker2022]_. The adaptive thermal heat balance (ATHB) framework
    introduced a method to account for the three adaptive principals, namely
    physiological, behavioral, and psychological adaptation, individually
    within existing heat balance models. The objective is a predictive model of
    thermal sensation applicable during the design stage or in international
    standards without knowing characteristics of future occupants.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, in [°C].
    tr : float or list of floats
        Mean radiant temperature, in [°C].
    vr : float or list of floats
        Relative air speed, in [m/s].

        .. note::
            vr is the relative air speed caused by body movement and not the air
            speed measured by the air speed sensor. The relative air speed is the sum of the
            average air speed measured by the sensor plus the activity-generated air speed
            (Vag). Where Vag is the activity-generated air speed caused by motion of
            individual body parts. vr can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.v_relative`.

    rh : float or list of floats
        Relative humidity, [%].
    met : float or list of floats
        Metabolic rate, [met].
    clo : bool, float or list of floats, optional
        Clothing insulation, in [clo]. If `clo` is set to False, the clothing insulation
        value will be calculated using the equation provided in the paper. Defaults to False.
    t_running_mean: float or list of floats
        Running mean temperature, in [°C].

        .. note::
            The running mean temperature can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.running_mean_outdoor_temperature`.

    Returns
    -------
    ATHBResults
        Dataclass containing the results of the ATHB calculation. See :py:class:`~pythermalcomfort.classes_return.ATHBResults` for more details.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import athb

        # calculate the predicted mean vote (PMV) using the Adaptive Thermal Heat Balance model
        results = athb(tdb=25, tr=25, vr=0.1, rh=50, met=1.2, t_running_mean=20)
        print(results.athb_pmv)  # returns the PMV value

        # for multiple points
        results = athb(
            tdb=[25, 25, 25],
            tr=[25, 25, 25],
            vr=[0.1, 0.1, 0.1],
            rh=[50, 50, 50],
            met=[1.2, 1.2, 1.2],
            t_running_mean=[20, 20, 20],
        )
        print(results.athb_pmv)

    """
    # Validate inputs using the ATHBInputs class
    ATHBInputs(tdb=tdb, tr=tr, vr=vr, rh=rh, met=met, t_running_mean=t_running_mean)

    tdb = np.array(tdb)
    tr = np.array(tr)
    vr = np.array(vr)
    met = np.array(met)
    rh = np.array(rh)
    t_running_mean = np.array(t_running_mean)

    met_adapted = met - (0.234 * t_running_mean) / 58.2

    clo_adapted = clo
    if clo is False:
        # Adapted clothing insulation level through behavioral adaptation
        clo_adapted = np.power(
            10,
            (
                -0.17168
                - 0.000485 * t_running_mean
                + 0.08176 * met_adapted
                - 0.00527 * t_running_mean * met_adapted
            ),
        )

    pmv_res = _pmv_ppd_optimized(tdb, tr, vr, rh, met_adapted, clo_adapted, 0)
    ts = 0.303 * np.exp(-0.036 * met_adapted * met_to_w_m2) + 0.028
    l_adapted = pmv_res / ts

    # Predicted thermal sensation vote
    athb_pmv = np.around(
        1.484
        + 0.0276 * l_adapted
        - 0.9602 * met_adapted
        - 0.0342 * t_running_mean
        + 0.0002264 * l_adapted * t_running_mean
        + 0.018696 * met_adapted * t_running_mean
        - 0.0002909 * l_adapted * met_adapted * t_running_mean,
        3,
    )

    return ATHB(athb_pmv=athb_pmv)



import numpy as np
import pytest

from pythermalcomfort.models import heat_index_rothfusz
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_heat_index(get_test_url, retrieve_data) -> None:
    """Test the heat index function with various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.HEAT_INDEX.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = heat_index_rothfusz(**inputs, limit_inputs=False)

        validate_result(result, outputs, tolerance)


def test_single_input_caution() -> None:
    """Test that the function returns a single value with caution category."""
    result = heat_index_rothfusz(tdb=29, rh=50, round_output=True)
    assert result.hi.shape == ()  # zero-dim ndarray
    assert result.hi.item() == pytest.approx(29.7, rel=1e-3)
    assert result.stress_category == "caution"


def test_below_threshold_produces_nan() -> None:
    """Test that the function returns NaN for heat index below threshold."""
    result = heat_index_rothfusz(tdb=25, rh=80)
    assert np.isnan(result.hi).item()
    assert isinstance(result.stress_category, np.str_)


def test_vector_input_no_rounding() -> None:
    """Test that the function handles vector inputs correctly without rounding."""
    tdb = [30.0, 20.0, 28.5]
    rh = [70.0, 90.0, 50.0]
    result = heat_index_rothfusz(tdb=tdb, rh=rh, round_output=False)
    hi = result.hi
    # Shape matches inputs
    assert hi.shape == (3,)
    # First element has multiple decimals and matches expected formula
    assert hi[0] == pytest.approx(35.33, rel=1e-2)
    # Second element below threshold ⇒ NaN
    assert np.isnan(hi[1])
    # Third element above threshold ⇒ finite number
    assert np.isfinite(hi[2])
    # Stress categories array matches hi length
    assert isinstance(result.stress_category, np.ndarray)
    assert result.stress_category.shape == (3,)
    assert result.stress_category[0] == "extreme caution"



from __future__ import annotations

import math

import numpy as np

from pythermalcomfort.classes_input import GaggeTwoNodesSleepInputs
from pythermalcomfort.classes_return import GaggeTwoNodesSleep


def two_nodes_gagge_sleep(
    tdb: float | list[float],
    tr: float | list[float],
    v: float | list[float],
    rh: float | list[float],
    clo: float | list[float],
    thickness_quilt: float | list[float],
    wme: float = 0,
    p_atm: float = 101325,
    **kwargs,
) -> GaggeTwoNodesSleep:
    """Adaption of the Gagge two-node model for sleep thermal environment, developed by Yan, S., Xiong, J., Kim, J. and de Dear, R. [Yan2022]_.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].

        .. note::
            tdb, tr, v, rh, clo and thickness must have the same length.
            This length will be the duration of the simulation.

    tr : float or list of floats
        Mean radiant temperature, [°C].
    v : float or list of floats
        Air speed, [m/s].
    rh : float or list of floats
        Relative humidity, [%].
    clo : float or list of floats
        Clothing insulation, [clo].
    thickness_quilt : float or list of floats
        Thickness of the quilt. [cm].
    wme : float, optional
        External work, [met]. Defaults to 0.
    p_atm : float, optional
        Atmospheric pressure, default value 101325 [Pa]. Defaults to 101325.
    **kwargs : dict
        Keyword arguments:

        * ltime : int, optional
            Number of time steps for each iteration. Defaults to 1.
        * height : float, optional
            Height of the person, [cm]. Defaults to 171.
        * weight : float, optional
            Weight of the person, [kg]. Defaults to 70.
        * c_sw : float, optional
            Driving coefficient for regulatory sweating. Defaults to 170.
        * c_dil : float, optional
            Driving coefficient for vasodilation. Defaults to 120.
        * c_str : float, optional
            Driving coefficient for vasoconstriction. Defaults to 0.5.
        * temp_skin_neutral : float, optional
            Skin temperature at neutral conditions, [°C]. Defaults to 33.7.
        * temp_core_neutral : float, optional
            Core temperature at neutral conditions, [°C]. Defaults to 36.8.
        * e_skin : float, optional
            Total evaporative heat loss, [W]. Defaults to 0.094.
        * alfa : float, optional
            Dynamic fraction of total body mass assigned to the skin node. Defaults to 0.1.
        * skin_blood_flow : float, optional
            Skin-blood-flow rate per unit surface area, [kg/h/m2]. Defaults to 6.3.
        * met_shivering : float, optional
            Metabolic rate due to shivering, [met]. Defaults to 0.

    Returns
    -------
    GaggeTwoNodesSleep
        A dataclass containing the results of the Gagge two-node model for sleep thermal environment.
        See :py:class:`~pythermalcomfort.classes_return.GaggeTwoNodesSleep` for more details.
        To access the results, use the corresponding attributes of the returned instance, e.g. `result.e_skin`.

    """
    ltime = kwargs.pop("ltime", 1)
    height = kwargs.pop("height", 171)
    weight = kwargs.pop("weight", 70)
    c_sw = kwargs.pop("c_sw", 170)
    c_dil = kwargs.pop("c_dil", 120)
    c_str = kwargs.pop("c_str", 0.5)
    temp_skin_neutral = kwargs.pop("temp_skin_neutral", 33.7)
    temp_core_neutral = kwargs.pop("temp_core_neutral", 36.8)
    e_skin = kwargs.pop("e_skin", 0.094)
    alfa = kwargs.pop("alfa", 0.1)
    skin_blood_flow = kwargs.pop("skin_blood_flow", 6.3)
    met_shivering = kwargs.pop("met_shivering", 0)

    if kwargs:
        error_msg = f"Unexpected keyword arguments: {list(kwargs.keys())}"
        raise TypeError(error_msg)

    GaggeTwoNodesSleepInputs(
        tdb=tdb,
        tr=tr,
        v=v,
        rh=rh,
        clo=clo,
        thickness_quilt=thickness_quilt,
        wme=wme,
        p_atm=p_atm,
    )

    tdb = np.atleast_1d(tdb)
    tr = np.atleast_1d(tr)
    v = np.atleast_1d(v)
    rh = np.atleast_1d(rh)
    clo = np.atleast_1d(clo)
    thickness_quilt = np.atleast_1d(thickness_quilt)

    # These variables should have the same length, which will be the duration
    lengths = [len(x) for x in (tdb, tr, v, rh, clo, thickness_quilt)]
    if len(set(lengths)) != 1:
        error_message = f"Parameters tdb, tr, v, rh, clo and thickness must have the same length. Got lengths {lengths}"
        raise ValueError(error_message)
    duration = lengths[0]

    # Initialize physiological state variables to be updated in each iteration
    result = {
        "t_core": temp_core_neutral,
        "t_skin": temp_skin_neutral,
        "e_skin": e_skin,
        "met_shivering": met_shivering,
        "alfa": alfa,
        "skin_blood_flow": skin_blood_flow,
    }

    results = []

    for i in range(duration):
        # Calculate metabolic rate using polynomial equation from Yan et al. (2022)
        met = (
            -0.000000000000575 * ((i - 1) / 60) ** 5
            + 0.000000000785521 * ((i - 1) / 60) ** 4
            - 0.00000039173563 * ((i - 1) / 60) ** 3
            + 0.000087620232151 * ((i - 1) / 60) ** 2
            - 0.008801558913211 * ((i - 1) / 60)
            + 1.09952538864493
        )

        # Calculate core temperature using quadratic equation from Yan et al. (2022)
        t_core = 0.022234 * ((i - 1) / 60) ** 2 - 0.27677 * ((i - 1) / 60) + 37.02

        result = _sleep_set(
            tdb[i],
            tr[i],
            v[i],
            rh[i],
            clo[i],
            thickness=thickness_quilt[i],
            met=met,
            wme=wme,
            p_atm=p_atm,
            ltime=ltime,
            height=height,
            weight=weight,
            c_sw=c_sw,
            c_dil=c_dil,
            c_str=c_str,
            temp_skin_neutral=result["t_skin"],
            temp_core_neutral=t_core,
            e_skin=result["e_skin"],
            alfa=result["alfa"],
            skin_blood_flow=result["skin_blood_flow"],
            met_shivering=result["met_shivering"],
        )

        # results should be a list of the local dataclass
        results.append(result)

    if not results:
        output = {}
    else:
        output = {}
        for key in results[0]:
            vals = [d[key] for d in results]
            # only wrap in an array if there’s more than one element
            output[key] = np.array(vals) if len(vals) > 1 else vals[0]

    return GaggeTwoNodesSleep(**output)


def _sleep_set(
    tdb: float,
    tr: float,
    v: float,
    rh: float,
    clo: float,
    thickness: float,
    met: float,
    wme: float,
    p_atm: float,
    ltime: int,
    height: float,
    weight: float,
    c_sw: float,
    c_dil: float,
    c_str: float,
    temp_skin_neutral: float,
    temp_core_neutral: float,
    e_skin: float,
    alfa: float,
    skin_blood_flow: float,
    met_shivering: float,
) -> dict:
    m = met * 58.2
    w = wme * 58.2
    k_clo = 0.25
    temp_body_neutral = 36.49
    skin_blood_flow_neutral = 6.3
    sbc = 5.6697 * 10**-8
    sa = ((height * weight) / 3600) ** 0.5

    v = max(v, 0.1)
    t_skin = temp_skin_neutral
    t_core = temp_core_neutral
    rmm = m

    # initialize some variables
    e_rsw = 0  # heat lost by vaporization sweat
    e_diff = 0  # vapor diffusion through skin
    e_max = 0  # maximum evaporative capacity
    dry = 0
    r_ea = 0
    r_ecl = 0
    p_wet = 0
    t_body = 0
    x = 0

    pressure_in_atmospheres = p_atm / 101325
    r_clo = 0.155 * clo
    f_a_cl = 0.0308 * thickness + 0.7695
    lr = 2.2 / pressure_in_atmospheres
    pa = rh * math.exp(18.6686 - (4030.183 / (tdb + 235))) / 100
    if clo <= 0:
        w_max = 0.38 * v**-0.29
        i_cl = 1
    else:
        w_max = 0.59 * v**-0.08
        i_cl = 0.45

    temp_diff = abs(t_skin - tdb)
    chc = 4.4854 * temp_diff**0.2740 * v**0.1242
    h_r = 3.235
    ctc = h_r + chc
    r_a = 1 / (f_a_cl * ctc)
    t_op = (h_r * tr + chc * tdb) / ctc
    t_cl = t_op + (t_skin - t_op) / (ctc * (r_a + r_clo))
    t_cl_old = t_cl
    flag = False

    for _ in range(ltime):
        if flag:
            t_cl = (r_a * t_skin + r_clo * t_op) / (r_a + r_clo)
            if abs(t_cl - t_cl_old) > 0.01:
                flag = False
                t_cl_old = t_cl
            else:
                flag = True

        max_iter = 100
        iter_cnt = 0
        while not flag and iter_cnt < max_iter:
            h_r = 4 * sbc * ((t_cl + tr) / 2 + 273.15) ** 3 * 0.72
            ctc = h_r + chc
            r_a = 1 / (f_a_cl * ctc)
            t_op = (h_r * tr + chc * tdb) / ctc
            t_cl = (r_a * t_skin + r_clo * t_op) / (r_a + r_clo)
            if abs(t_cl - t_cl_old) > 0.01:
                flag = False
                t_cl_old = t_cl
            else:
                flag = True
            iter_cnt += 1

        dry = (t_skin - t_op) / (r_a + r_clo)
        hf_cs = (t_core - t_skin) * (5.28 + 1.163 * skin_blood_flow)
        q_res = 0.0023 * m * (44 - pa)
        c_res = 0.0014 * m * (34 - tdb)
        s_core = m - hf_cs - q_res - c_res - w
        s_skin = hf_cs - dry - e_skin
        tc_sk = 0.97 * alfa * weight
        tc_cr = 0.97 * (1 - alfa) * weight
        d_t_sk = (s_skin * sa) / tc_sk / 60
        d_t_cr = (s_core * sa) / tc_cr / 60
        t_skin += d_t_sk
        t_core += d_t_cr
        t_body = alfa * t_skin + (1 - alfa) * t_core

        warm_sk = max(t_skin - 33.7, 0)
        cold_s = max(33.7 - t_skin, 0)
        warm_c = max(t_core - temp_core_neutral, 0)
        cold_c = max(temp_core_neutral - t_core, 0)
        warm_b = max(t_body - temp_body_neutral, 0)

        skin_blood_flow = (skin_blood_flow_neutral + c_dil * warm_c) / (
            1 + c_str * cold_s
        )
        skin_blood_flow = min(max(skin_blood_flow, 0.5), 90)
        reg_sw = c_sw * warm_b * math.exp(warm_sk / 10.7)
        reg_sw = min(reg_sw, 500)
        e_rsw = 0.68 * reg_sw
        r_ea = 1 / (lr * f_a_cl * chc)
        r_ecl = r_clo / (lr * i_cl)
        e_max = (_fnsvp(t_skin) - pa) / (r_ea + r_ecl)
        p_rsw = e_rsw / e_max
        p_wet = 0.06 + 0.94 * p_rsw
        e_diff = p_wet * e_max - e_rsw
        if p_wet > w_max:
            p_wet = w_max
            p_rsw = w_max / 0.94
            e_rsw = p_rsw * e_max
            e_diff = 0.06 * (1 - p_rsw) * e_max
        e_skin = e_rsw + e_diff
        met_shivering = 19.4 * cold_s * cold_c
        m = rmm + met_shivering
        alfa = 0.0417737 + 0.7451833 / (skin_blood_flow + 0.585417)

    q_skin = dry + e_skin
    rn = m - w
    e_comfort = 0.42 * (rn - 58.2)
    e_comfort = max(e_comfort, 0)
    e_max *= w_max
    h_d = 1 / (r_a + r_clo)
    h_e = 1 / (r_ea + r_ecl)
    wet = p_wet
    p_s_sk = _fnsvp(t_skin)
    chrs = h_r
    chcs = max(3, 5.66 * ((met - 0.85) ** 0.39) if met >= 0.85 else 0)
    ctcs = chcs + chrs
    r_clo_s = 1.52 / ((met - wme) + 0.6944) - 0.1835
    r_cl_s = 0.155 * r_clo_s
    f_a_cl_s = 1 + k_clo * r_clo_s
    f_cl_s = 1 / (1 + 0.155 * f_a_cl_s * ctcs * r_clo_s)
    i_m_s = 0.45
    i_cl_s = i_m_s * chcs / ctcs * (1 - f_cl_s) / (chcs / ctcs - f_cl_s * i_m_s)
    r_a_s = 1 / (f_a_cl_s * ctcs)
    r_ea_s = 1 / (lr * f_a_cl_s * chcs)
    r_ecl_s = r_cl_s / (lr * i_cl_s)
    h_d_s = 1 / (r_a_s + r_cl_s)
    h_e_s = 1 / (r_ea_s + r_ecl_s)
    delta = 1e-04
    xold = t_skin - q_skin / h_d

    flag1 = False
    while not flag1:
        err1 = _fnerre(xold, q_skin, h_d, t_skin, wet, h_e, p_s_sk)
        err2 = _fnerre(xold + delta, q_skin, h_d, t_skin, wet, h_e, p_s_sk)
        err_diff = err2 - err1
        if abs(err_diff) < 1e-10:  # Avoid division by very small values
            # Use a fallback approach or break iteration
            break
        x = xold - delta * err1 / err_diff
        if abs(x - xold) > 0.01:
            xold = x
            flag1 = False
        else:
            flag1 = True

    xold = t_skin - q_skin / h_d_s

    flag2 = False
    while not flag2:
        err1 = _fnerrs(xold, q_skin, h_d_s, t_skin, wet, h_e_s, p_s_sk)
        err2 = _fnerrs(xold + delta, q_skin, h_d_s, t_skin, wet, h_e_s, p_s_sk)
        x = xold - delta * err1 / (err2 - err1)
        if abs(x - xold) > 0.01:
            xold = x
            flag2 = False
        else:
            flag2 = True

    set_temp = x
    tbm_l = (0.194 / 58.15) * rn + 36.301
    tbm_h = (0.347 / 58.15) * rn + 36.669
    if t_body < tbm_l:
        t_sens = 0.4685 * (t_body - tbm_l)
    elif tbm_l <= t_body < tbm_h:
        t_sens = w_max * 4.7 * (t_body - tbm_l) / (tbm_h - tbm_l)
    else:
        t_sens = w_max * 4.7 + 0.4685 * (t_body - tbm_h)

    disc = 4.7 * (e_rsw - e_comfort) / (e_max - e_comfort - e_diff)
    if disc < 0:
        disc = t_sens

    return {
        "set": set_temp,
        "t_core": t_core,
        "t_skin": t_skin,
        "wet": wet,
        "t_sens": t_sens,
        "disc": disc,
        "e_skin": e_skin,
        "met_shivering": met_shivering,
        "alfa": alfa,
        "skin_blood_flow": skin_blood_flow,
    }


def _fnsvp(t):
    """Calculate saturation vapor pressure at temperature t.

    Parameters
    ----------
    t : float
        Temperature [°C]

    Returns
    -------
    float
        Saturation vapor pressure [Pa]

    """
    return math.exp(18.6686 - 4030.183 / (t + 235))


def _fnerre(x, hsk, hd, tsk, w, he, pssk):
    """Error function for iterative solution of SET temperature.

    Used in the Newton-Raphson algorithm.
    """
    return hsk - hd * (tsk - x) - w * he * (pssk - 0.5 * _fnsvp(x))


def _fnerrs(x, hsk, hd_s, tsk, w, he_s, pssk):
    """Error function for iterative solution of SET temperature (second version).

    Used in the Newton-Raphson algorithm.
    """
    return hsk - hd_s * (tsk - x) - w * he_s * (pssk - 0.5 * _fnsvp(x))



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import THIInputs
from pythermalcomfort.classes_return import THI


def thi(
    tdb: float | list[float],
    rh: float | list[float],
    round_output: bool = True,
) -> THI:
    """Calculate the Temperature-Humidity Index (THI) defined in [Yan2025]_, equivalent to
    the definition in [Schlatter1987]_, but uses Celsius instead of Fahrenheit.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    rh: float or list of floats
        Relative humidity, [%].
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    THI
        A dataclass containing the Temperature-Humidity Index.
        See :py:class:`~pythermalcomfort.classes_return.THI` for more details.
        To access the `thi` value, use the `thi` attribute of the returned `THI`
        instance, e.g., `result.thi`.

    """
    # Validate inputs using the THIInputs class
    THIInputs(
        tdb=tdb,
        rh=rh,
        round_output=round_output,
    )

    tdb = np.array(tdb)
    rh = np.array(rh)

    _thi = 1.8 * tdb + 32 - 0.55 * (1 - 0.01 * rh) * (1.8 * tdb - 26)

    if round_output:
        _thi = np.round(_thi, 1)

    return THI(thi=_thi)
